from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.custom_codec import *
from hazelcast.protocol.codec.client_message_type import *
from hazelcast.protocol.event_response_const import *
from hazelcast.six.moves import range

REQUEST_TYPE = CLIENT_ADDMEMBERSHIPLISTENER
RESPONSE_TYPE = 104
RETRYABLE = False


def calculate_size(local_only):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += BOOLEAN_SIZE_IN_BYTES
    return data_size


def encode_request(local_only):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(local_only))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_bool(local_only)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(response=None)
    parameters['response'] = client_message.read_str()
    return parameters


def handle(client_message, handle_event_member=None, handle_event_member_list=None, handle_event_member_attribute_change=None, to_object=None):
    """ Event handler """
    message_type = client_message.get_message_type()
    if message_type == EVENT_MEMBER and handle_event_member is not None:
        member = MemberCodec.decode(client_message, to_object)
        event_type = client_message.read_int()
        handle_event_member(member=member, event_type=event_type)
    if message_type == EVENT_MEMBERLIST and handle_event_member_list is not None:
        members_size = client_message.read_int()
        members = []
        for _ in range(0, members_size):
            members_item = MemberCodec.decode(client_message, to_object)
            members.append(members_item)
        handle_event_member_list(members=members)
    if message_type == EVENT_MEMBERATTRIBUTECHANGE and handle_event_member_attribute_change is not None:
        uuid = client_message.read_str()
        key = client_message.read_str()
        operation_type = client_message.read_int()
        value = None
        if not client_message.read_bool():
            value = client_message.read_str()
        handle_event_member_attribute_change(uuid=uuid, key=key, operation_type=operation_type, value=value)



from struct import unpack_from

from hazelcast.hash import murmur_hash3_x86_32
from hazelcast.serialization.bits import *
from hazelcast.serialization.serialization_const import *

PARTITION_HASH_OFFSET = 0
TYPE_OFFSET = 4
DATA_OFFSET = 8

HEAP_DATA_OVERHEAD = DATA_OFFSET


class Data(object):
    """
    Data is basic unit of serialization. It stores binary form of an object serialized by serialization service
    """

    def __init__(self, buff=None):
        self._buffer = buff

    def to_bytes(self):
        """
        Returns byte array representation of internal binary format.

        :return:  (byte array), byte array representation of internal binary format.
        """
        return self._buffer

    def get_type(self):
        """
        Returns serialization type of binary form.

        :return: Serialization type of binary form.
        """
        if self.total_size() == 0:
            return CONSTANT_TYPE_NULL
        return unpack_from(FMT_BE_INT, self._buffer, TYPE_OFFSET)[0]

    def total_size(self):
        """
        Returns the total size of Data in bytes.

        :return: (int), total size of Data in bytes.
        """
        return len(self._buffer) if self._buffer is not None else 0

    def data_size(self):
        """
        Returns size of internal binary data in bytes.

        :return: (int), size of internal binary data in bytes.
        """
        return max(self.total_size() - HEAP_DATA_OVERHEAD, 0)

    def get_partition_hash(self):
        """
        Returns partition hash calculated for serialized object.
            Partition hash is used to determine partition of a Data and is calculated using
                * PartitioningStrategy during serialization.
                * If partition hash is not set then hash_code() is used.

        :return: partition hash
        """
        if self.has_partition_hash():
            return unpack_from(FMT_BE_INT, self._buffer, PARTITION_HASH_OFFSET)[0]
        return self.hash_code()

    def is_portable(self):
        """
        Determines whether this Data is created from a :class:`~hazelcast.serialization.api.Portable`. object or not.

        :return: (bool), ``true`` if source object is Portable, ``false`` otherwise.
        """
        return CONSTANT_TYPE_PORTABLE == self.get_type()

    def has_partition_hash(self):
        """
        Determines whether this Data has partition hash or not.

        :return: (bool), ``true`` if Data has partition hash, ``false`` otherwise.
        """
        return self._buffer is not None \
               and len(self._buffer) >= HEAP_DATA_OVERHEAD \
               and unpack_from(FMT_BE_INT, self._buffer, PARTITION_HASH_OFFSET)[0] != 0

    def hash_code(self):
        """
        Returns the murmur hash of the internal data.

        :return: the murmur hash of the internal data.
        """
        return murmur_hash3_x86_32(self._buffer, DATA_OFFSET, self.data_size())

    def __hash__(self):
        return self.hash_code()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.total_size() == other.total_size() and self._buffer == other.to_bytes()

    def __len__(self):
        return self.total_size()




from hazelcast.protocol.codec import atomic_long_add_and_get_codec, atomic_long_compare_and_set_codec, \
    atomic_long_decrement_and_get_codec, atomic_long_get_and_add_codec, atomic_long_get_and_increment_codec, \
    atomic_long_get_and_set_codec, atomic_long_get_codec, atomic_long_increment_and_get_codec, atomic_long_set_codec, \
    atomic_long_alter_and_get_codec, atomic_long_alter_codec, atomic_long_apply_codec, atomic_long_get_and_alter_codec
from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import check_not_none


class AtomicLong(PartitionSpecificProxy):
    """
    AtomicLong is a redundant and highly available distributed long value which can be updated atomically.
    """
    def add_and_get(self, delta):
        """
        Atomically adds the given delta value to the currently stored value.

        :param delta: (long), the value to add to the currently stored value.
        :return: (long), the updated value.
        """
        return self._encode_invoke(atomic_long_add_and_get_codec, delta=delta)

    def alter(self, function):
        """
        Alters the currently stored value by applying a function on it.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_long_alter_codec, function=self._to_data(function))

    def alter_and_get(self, function):
        """
        Alters the currently stored value by applying a function on it and gets the result.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (long), the new value.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_long_alter_and_get_codec, function=self._to_data(function))

    def apply(self, function):
        """
        Applies a function on the value, the actual stored value will not change.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (object), the result of the function application.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_long_apply_codec, function=self._to_data(function))

    def compare_and_set(self, expected, updated):
        """
        Atomically sets the value to the given updated value only if the current value == the expected value.

        :param expected: (long), the expected value.
        :param updated: (long), the new value.
        :return: (bool), ``true`` if successful; or ``false`` if the actual value was not equal to the expected value.
        """
        return self._encode_invoke(atomic_long_compare_and_set_codec, expected=expected,
                                   updated=updated)

    def decrement_and_get(self):
        """
        Atomically decrements the current value by one.

        :return: (long), the updated value, the current value decremented by one.
        """
        return self._encode_invoke(atomic_long_decrement_and_get_codec)

    def get(self):
        """
        Gets the current value.

        :return: (long), gets the current value.
        """
        return self._encode_invoke(atomic_long_get_codec)

    def get_and_add(self, delta):
        """
        Atomically adds the given value to the current value.

        :param delta: (long), the value to add to the current value.
        :return: (long), the old value before the addition.
        """
        return self._encode_invoke(atomic_long_get_and_add_codec, delta=delta)

    def get_and_alter(self, function):
        """
        Alters the currently stored value by applying a function on it on and gets the old value.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (long), the old value.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_long_get_and_alter_codec, function=self._to_data(function))

    def get_and_set(self, new_value):
        """
        Atomically sets the given value and returns the old value.

        :param new_value: (long), the new value.
        :return: (long), the old value.
        """
        return self._encode_invoke(atomic_long_get_and_set_codec, new_value=new_value)

    def increment_and_get(self):
        """
        Atomically increments the current value by one.

        :return: (long), the updated value, the current value incremented by one.
        """
        return self._encode_invoke(atomic_long_increment_and_get_codec)

    def get_and_increment(self):
        """
        Atomically increments the current value by one.

        :return: (long), the old value.
        """
        return self._encode_invoke(atomic_long_get_and_increment_codec)

    def set(self, new_value):
        """
        Atomically sets the given value.

        :param new_value: (long), the new value.
        """
        return self._encode_invoke(atomic_long_set_codec, new_value=new_value)




TRANSACTIONALMULTIMAP_PUT = 0x1101
TRANSACTIONALMULTIMAP_GET = 0x1102
TRANSACTIONALMULTIMAP_REMOVE = 0x1103
TRANSACTIONALMULTIMAP_REMOVEENTRY = 0x1104
TRANSACTIONALMULTIMAP_VALUECOUNT = 0x1105
TRANSACTIONALMULTIMAP_SIZE = 0x1106



import json
import os
import subprocess


SERIALIZATION_VERSION = 1
CLIENT_TYPE = "PYH"
CLIENT_VERSION_INFO = (3, 12, 2)
CLIENT_VERSION = ".".join(map(str, CLIENT_VERSION_INFO))


class GitInfoProvider(object):
    def __init__(self):
        here = os.path.abspath(os.path.dirname(__file__))
        git_info = read_git_info(here)

        if git_info:
            self.commit_id = git_info.get("commit_id", "")
            self.commit_date = git_info.get("commit_date", "")
        else:
            self.commit_id = get_commit_id()
            self.commit_date = get_commit_date()
            write_git_info(here, self.commit_id, self.commit_date)


def read_git_info(here):
    try:
        file_path = os.path.abspath(os.path.join(here, "git_info.json"))
        with open(file_path, "r") as f:
            info = json.load(f)
            return info
    except:
        return None


def write_git_info(here, commit_id, commit_date):
    try:
        git_info = dict()
        git_info["commit_id"] = commit_id
        git_info["commit_date"] = commit_date

        file_path = os.path.abspath(os.path.join(here, "git_info.json"))
        with open(file_path, "w") as f:
            json.dump(git_info, f)
    except:
        pass


def get_commit_id():
    try:
        commit_id = subprocess.check_output(["git", "show", "-s", "--format=\"%h\""]).decode()
        commit_id = commit_id.strip().replace("\"", "").replace("'", "")
        return commit_id
    except:
        return ""


def get_commit_date():
    try:
        commit_date = subprocess.check_output(["git", "show", "-s", "--format=\"%cd\"", "--date=short"]).decode()
        commit_date = commit_date.strip().replace("\"", "").replace("'", "").replace("-", "")
        return commit_date
    except:
        return ""


_info_provider = GitInfoProvider()
GIT_COMMIT_ID = _info_provider.commit_id
GIT_COMMIT_DATE = _info_provider.commit_date



from hazelcast.exception import HazelcastSerializationError
from hazelcast.util import enum
from hazelcast import six

FieldType = enum(
        PORTABLE=0,
        BYTE=1,
        BOOLEAN=2,
        CHAR=3,
        SHORT=4,
        INT=5,
        LONG=6,
        FLOAT=7,
        DOUBLE=8,
        UTF=9,
        PORTABLE_ARRAY=10,
        BYTE_ARRAY=11,
        BOOLEAN_ARRAY=12,
        CHAR_ARRAY=13,
        SHORT_ARRAY=14,
        INT_ARRAY=15,
        LONG_ARRAY=16,
        FLOAT_ARRAY=17,
        DOUBLE_ARRAY=18,
        UTF_ARRAY=19
)


class FieldDefinition(object):
    def __init__(self, index, field_name, field_type, version, factory_id=0, class_id=0):
        self.index = index
        self.field_name = field_name
        self.field_type = field_type
        self.version = version
        self.factory_id = factory_id
        self.class_id = class_id

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and (self.index, self.field_name, self.field_type, self.version, self.factory_id, self.class_id) == \
                   (other.index, other.field_name, other.field_type, other.version, other.factory_id, other.class_id)

    def __repr__(self):
        return "FieldDefinition[ ix:{}, name:{}, type:{}, version:{}, fid:{}, cid:{}]".format(self.index,
                                                                                              self.field_name,
                                                                                              self.field_type,
                                                                                              self.version,
                                                                                              self.factory_id,
                                                                                              self.class_id)


class ClassDefinition(object):
    def __init__(self, factory_id, class_id, version):
        self.factory_id = factory_id
        self.class_id = class_id
        self.version = version
        self.field_defs = {}  # string:FieldDefinition

    def add_field_def(self, field_def):
        self.field_defs[field_def.field_name] = field_def

    def get_field(self, field_name_or_index):
        if isinstance(field_name_or_index, int):
            index = field_name_or_index
            count = self.get_field_count()
            if 0 <= index < count:
                for field in six.itervalues(self.field_defs):
                    if field.index == index:
                        return field
            raise IndexError("Index is out of bound. Index: {} and size: {}".format(index, count))
        else:
            return self.field_defs.get(field_name_or_index, None)

    def has_field(self, field_name):
        return field_name in self.field_defs

    def get_field_names(self):
        return list(self.field_defs.keys())

    def get_field_type(self, field_name):
        fd = self.get_field(field_name)
        if fd:
            return fd.field_type
        raise ValueError("Unknown field: {}".format(field_name))

    def get_field_class_id(self, field_name):
        fd = self.get_field(field_name)
        if fd:
            return fd.class_id
        raise ValueError("Unknown field: {}".format(field_name))

    def get_field_count(self):
        return len(self.field_defs)

    def set_version_if_not_set(self, version):
        if self.version < 0:
            self.version = version

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.factory_id, self.class_id, self.version, self.field_defs) == \
                                                     (other.factory_id, other.class_id, other.version, other.field_defs)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "fid:{}, cid:{}, v:{}, fields:{}".format(self.factory_id, self.class_id, self.version, self.field_defs)

    def __hash__(self):
        return id(self)//16


class ClassDefinitionBuilder(object):
    def __init__(self, factory_id, class_id, version=0):
        self.factory_id = factory_id
        self.class_id = class_id
        self.version = version

        self._index = 0
        self._done = False
        self._field_defs = list()

    def add_portable_field(self, field_name, class_def):
        if class_def.class_id is None or class_def.class_id == 0:
            raise ValueError("Portable class id cannot be zero!")
        self._add_field_by_type(field_name, FieldType.PORTABLE, class_def.version,
                                class_def.factory_id, class_def.class_id)
        return self

    def add_byte_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.BYTE, self.version)
        return self

    def add_boolean_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.BOOLEAN, self.version)
        return self

    def add_char_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.CHAR, self.version)
        return self

    def add_short_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.SHORT, self.version)
        return self

    def add_int_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.INT, self.version)
        return self

    def add_long_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.LONG, self.version)
        return self

    def add_float_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.FLOAT, self.version)
        return self

    def add_double_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.DOUBLE, self.version)
        return self

    def add_utf_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.UTF, self.version)
        return self

    def add_portable_array_field(self, field_name, class_def):
        if class_def.class_id is None or class_def.class_id == 0:
            raise ValueError("Portable class id cannot be zero!")
        self._add_field_by_type(field_name, FieldType.PORTABLE_ARRAY, class_def.version,
                                class_def.factory_id, class_def.class_id)
        return self

    def add_byte_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.BYTE_ARRAY, self.version)
        return self

    def add_boolean_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.BOOLEAN_ARRAY, self.version)
        return self

    def add_char_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.CHAR_ARRAY, self.version)
        return self

    def add_short_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.SHORT_ARRAY, self.version)
        return self

    def add_int_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.INT_ARRAY, self.version)
        return self

    def add_long_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.LONG_ARRAY, self.version)
        return self

    def add_float_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.FLOAT_ARRAY, self.version)
        return self

    def add_double_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.DOUBLE_ARRAY, self.version)
        return self

    def add_utf_array_field(self, field_name):
        self._add_field_by_type(field_name, FieldType.UTF_ARRAY, self.version)
        return self

    def add_field_def(self, field_def):
        self._check()
        if self._index != field_def.index:
            raise ValueError("Invalid field index")
        self._index += 1
        self._field_defs.append(field_def)
        return self

    def build(self):
        self._done = True
        cd = ClassDefinition(self.factory_id, self.class_id, self.version)
        for field_def in self._field_defs:
            cd.add_field_def(field_def)
        return cd

    def _add_field_by_type(self, field_name, field_type, version, factory_id=0, class_id=0):
        self._check()
        self._field_defs.append(FieldDefinition(self._index, field_name, field_type, version, factory_id, class_id))
        self._index += 1

    def _check(self):
        if self._done:
            raise HazelcastSerializationError("ClassDefinition is already built for {}".format(self.class_id))



"""
Each exception that are defined in client protocol have unique identifier which are error code.
All error codes defined in protocol are listed in this class.
"""

# Message type indicating exception
EXCEPTION_MESSAGE_TYPE = 109

UNDEFINED = 0
ARRAY_INDEX_OUT_OF_BOUNDS = 1
ARRAY_STORE = 2
AUTHENTICATION = 3
CACHE = 4
CACHE_LOADER = 5
CACHE_NOT_EXISTS = 6
CACHE_WRITER = 7
CALLER_NOT_MEMBER = 8
CANCELLATION = 9
CLASS_CAST = 10
CLASS_NOT_FOUND = 11
CONCURRENT_MODIFICATION = 12
CONFIG_MISMATCH = 13
CONFIGURATION = 14
DISTRIBUTED_OBJECT_DESTROYED = 15
DUPLICATE_INSTANCE_NAME = 16
EOF = 17
ENTRY_PROCESSOR = 18
EXECUTION = 19
HAZELCAST = 20
HAZELCAST_INSTANCE_NOT_ACTIVE = 21
HAZELCAST_OVERLOAD = 22
HAZELCAST_SERIALIZATION = 23
IO = 24
ILLEGAL_ARGUMENT = 25
ILLEGAL_ACCESS_EXCEPTION = 26
ILLEGAL_ACCESS_ERROR = 27
ILLEGAL_MONITOR_STATE = 28
ILLEGAL_STATE = 29
ILLEGAL_THREAD_STATE = 30
INDEX_OUT_OF_BOUNDS = 31
INTERRUPTED = 32
INVALID_ADDRESS = 33
INVALID_CONFIGURATION = 34
MEMBER_LEFT = 35
NEGATIVE_ARRAY_SIZE = 36
NO_SUCH_ELEMENT = 37
NOT_SERIALIZABLE = 38
NULL_POINTER = 39
OPERATION_TIMEOUT = 40
PARTITION_MIGRATING = 41
QUERY = 42
QUERY_RESULT_SIZE_EXCEEDED = 43
QUORUM = 44
REACHED_MAX_SIZE = 45
REJECTED_EXECUTION = 46
REMOTE_MAP_REDUCE = 47
RESPONSE_ALREADY_SENT = 48
RETRYABLE_HAZELCAST = 49
RETRYABLE_IO = 50
RUNTIME = 51
SECURITY = 52
SOCKET = 53
STALE_SEQUENCE = 54
TARGET_DISCONNECTED = 55
TARGET_NOT_MEMBER = 56
TIMEOUT = 57
TOPIC_OVERLOAD = 58
TOPOLOGY_CHANGED = 59
TRANSACTION = 60
TRANSACTION_NOT_ACTIVE = 61
TRANSACTION_TIMED_OUT = 62
URI_SYNTAX = 63
UTF_DATA_FORMAT = 64
UNSUPPORTED_OPERATION = 65
WRONG_TARGET = 66
XA = 67
ACCESS_CONTROL = 68
LOGIN = 69
UNSUPPORTED_CALLBACK = 70
NO_DATA_MEMBER = 71
REPLICATED_MAP_CANT_BE_CREATED = 72
MAX_MESSAGE_SIZE_EXCEEDED = 73
WAN_REPLICATION_QUEUE_FULL = 74
ASSERTION_ERROR = 75
OUT_OF_MEMORY_ERROR = 76
STACK_OVERFLOW_ERROR = 77
NATIVE_OUT_OF_MEMORY_ERROR = 78
SERVICE_NOT_FOUND = 79
STALE_TASK_ID = 80
DUPLICATE_TASK = 81
STALE_TASK = 82
LOCAL_MEMBER_RESET = 83
INDETERMINATE_OPERATION_STATE = 84
FLAKE_ID_NODE_ID_OUT_OF_RANGE_EXCEPTION = 85
TARGET_NOT_REPLICA_EXCEPTION = 86
MUTATION_DISALLOWED_EXCEPTION = 87
CONSISTENCY_LOST_EXCEPTION = 88




ATOMICREFERENCE_APPLY = 0x0b01
ATOMICREFERENCE_ALTER = 0x0b02
ATOMICREFERENCE_ALTERANDGET = 0x0b03
ATOMICREFERENCE_GETANDALTER = 0x0b04
ATOMICREFERENCE_CONTAINS = 0x0b05
ATOMICREFERENCE_COMPAREANDSET = 0x0b06
ATOMICREFERENCE_GET = 0x0b08
ATOMICREFERENCE_SET = 0x0b09
ATOMICREFERENCE_CLEAR = 0x0b0a
ATOMICREFERENCE_GETANDSET = 0x0b0b
ATOMICREFERENCE_SETANDGET = 0x0b0c
ATOMICREFERENCE_ISNULL = 0x0b0d




QUEUE_OFFER = 0x0301
QUEUE_PUT = 0x0302
QUEUE_SIZE = 0x0303
QUEUE_REMOVE = 0x0304
QUEUE_POLL = 0x0305
QUEUE_TAKE = 0x0306
QUEUE_PEEK = 0x0307
QUEUE_ITERATOR = 0x0308
QUEUE_DRAINTO = 0x0309
QUEUE_DRAINTOMAXSIZE = 0x030a
QUEUE_CONTAINS = 0x030b
QUEUE_CONTAINSALL = 0x030c
QUEUE_COMPAREANDREMOVEALL = 0x030d
QUEUE_COMPAREANDRETAINALL = 0x030e
QUEUE_CLEAR = 0x030f
QUEUE_ADDALL = 0x0310
QUEUE_ADDLISTENER = 0x0311
QUEUE_REMOVELISTENER = 0x0312
QUEUE_REMAININGCAPACITY = 0x0313
QUEUE_ISEMPTY = 0x0314



import threading

from hazelcast import util
from hazelcast.exception import HazelcastSerializationError
from hazelcast.serialization import bits
from hazelcast.serialization.portable.classdef import ClassDefinition, ClassDefinitionBuilder, FieldType, FieldDefinition
from hazelcast.serialization.portable.writer import ClassDefinitionWriter
from hazelcast.six.moves import range


class PortableContext(object):
    def __init__(self, serialization_service, portable_version):
        self.serialization_service = serialization_service
        self.portable_version = portable_version
        self._class_defs = dict()  # {factory_id:ClassDefinitionContext}

    def get_class_version(self, factory_id, class_id):
        return self._get_class_def_context(factory_id).get_class_version(class_id)

    def set_class_version(self, factory_id, class_id, version):
        self._get_class_def_context(factory_id).set_class_version(class_id, version)

    def lookup_class_definition(self, factory_id, class_id, version):
        return self._get_class_def_context(factory_id).lookup(class_id, version)

    def read_class_definition(self, data_in, factory_id, class_id, version):
        register = True
        builder = ClassDefinitionBuilder(factory_id, class_id, version)

        # final position after portable is read
        data_in.read_int()

        # field count
        field_count = data_in.read_int()
        offset = data_in.position()
        for i in range(0, field_count):
            pos = data_in.read_int(offset + i * bits.INT_SIZE_IN_BYTES)
            data_in.set_position(pos)

            _len = data_in.read_short()
            field_name = bytearray(_len)
            data_in.read_into(field_name)

            field_type = data_in.read_byte()

            field_factory_id = 0
            field_class_id = 0
            field_version = version
            if field_type == FieldType.PORTABLE:
                # is null
                if data_in.read_boolean():
                    register = False
                field_factory_id = data_in.read_int()
                field_class_id = data_in.read_int()

                # TODO: what there's a null inner Portable field
                if register:
                    field_version = data_in.read_int()
                    self.read_class_definition(data_in, field_factory_id, field_class_id, field_version)
            elif field_type == FieldType.PORTABLE_ARRAY:
                k = data_in.read_int()
                field_factory_id = data_in.read_int()
                field_class_id = data_in.read_int()

                # TODO: what there's a null inner Portable field
                if k > 0:
                    p = data_in.read_int()
                    data_in.set_position(p)
                    field_version = data_in.read_int()
                    self.read_class_definition(data_in, field_factory_id, field_class_id, field_version)
                else:
                    register = False
            builder.add_field_def(FieldDefinition(i, field_name.decode('ascii'), field_type, field_version,
                                                  field_factory_id, field_class_id))
        class_def = builder.build()
        if register:
            class_def = self.register_class_definition(class_def)
        return class_def

    def register_class_definition(self, class_definition):
        return self._get_class_def_context(class_definition.factory_id).register(class_definition)

    def lookup_or_register_class_definition(self, portable):
        fid = portable.get_factory_id()
        cid = portable.get_class_id()
        portable_version = util.get_portable_version(portable, self.portable_version)
        class_def = self.lookup_class_definition(fid, cid, portable_version)

        if class_def is None:
            writer = ClassDefinitionWriter(self, fid, cid, portable_version)
            portable.write_portable(writer)
            class_def = writer.register_and_get()

        return class_def

    def _get_class_def_context(self, factory_id):
        if factory_id not in self._class_defs:
            self._class_defs[factory_id] = ClassDefinitionContext(factory_id, self.portable_version)
        return self._class_defs[factory_id]


class ClassDefinitionContext(object):
    def __init__(self, factory_id, portable_version):
        self._factory_id = factory_id
        self._portable_version = portable_version
        self._versioned_definitions = {}  # (class_id, version) : ClassDefinition
        self._current_class_versions = {}  # class_id:version
        self._lock = threading.RLock()

    def get_class_version(self, class_id):
        return self._current_class_versions.get(class_id, -1)

    def set_class_version(self, class_id, version):
        try:
            current_version = self._current_class_versions[class_id]
            if current_version != version:
                raise ValueError("Class-id: {} is already registered!".format(class_id))
        except KeyError:
            self._current_class_versions[class_id] = version

    def lookup(self, class_id, version):
        return self._versioned_definitions.get((class_id, version), None)

    def register(self, class_def):
        with self._lock:
            if class_def is None:
                return None
            if class_def.factory_id != self._factory_id:
                raise HazelcastSerializationError("Invalid factory-id! {} -> {}".format(self._factory_id, class_def))
            if isinstance(class_def, ClassDefinition):
                class_def.set_version_if_not_set(self._portable_version)
            combined_key = (class_def.class_id, class_def.version)
            if combined_key not in self._versioned_definitions:
                self._versioned_definitions[combined_key] = class_def
                return class_def
            current_class_def = self._versioned_definitions[combined_key]
            if isinstance(current_class_def, ClassDefinition):
                if current_class_def != class_def:
                    raise HazelcastSerializationError("Incompatible class-definitions with same class-id: {} vs {}"
                                                      .format(class_def, current_class_def))
                return current_class_def
            self._versioned_definitions[combined_key] = class_def
            return class_def



from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.custom_codec import *
from hazelcast.protocol.codec.client_message_type import *
from hazelcast.six.moves import range

REQUEST_TYPE = CLIENT_GETPARTITIONS
RESPONSE_TYPE = 108
RETRYABLE = False


def calculate_size():
    """ Calculates the request payload size"""
    data_size = 0
    return data_size


def encode_request():
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size())
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(partitions=None)
    partitions_size = client_message.read_int()
    partitions = {}
    for _ in range(0, partitions_size):
        partitions_key = AddressCodec.decode(client_message, to_object)
        partitions_val_size = client_message.read_int()
        partitions_val = []
        for _ in range(0, partitions_val_size):
            partitions_val_item = client_message.read_int()
            partitions_val.append(partitions_val_item)
        partitions[partitions_key] = partitions_val
    parameters['partitions'] = partitions
    return parameters



from hazelcast.protocol.codec import client_create_proxy_codec, client_destroy_proxy_codec
from hazelcast.proxy.atomic_long import AtomicLong
from hazelcast.proxy.atomic_reference import AtomicReference
from hazelcast.proxy.count_down_latch import CountDownLatch
from hazelcast.proxy.executor import Executor
from hazelcast.proxy.id_generator import IdGenerator
from hazelcast.proxy.list import List
from hazelcast.proxy.lock import Lock
from hazelcast.proxy.map import create_map_proxy
from hazelcast.proxy.multi_map import MultiMap
from hazelcast.proxy.queue import Queue
from hazelcast.proxy.reliable_topic import ReliableTopic
from hazelcast.proxy.replicated_map import ReplicatedMap
from hazelcast.proxy.ringbuffer import Ringbuffer
from hazelcast.proxy.semaphore import Semaphore
from hazelcast.proxy.set import Set
from hazelcast.proxy.topic import Topic
from hazelcast.proxy.pn_counter import PNCounter
from hazelcast.proxy.flake_id_generator import FlakeIdGenerator

ATOMIC_LONG_SERVICE = "hz:impl:atomicLongService"
ATOMIC_REFERENCE_SERVICE = "hz:impl:atomicReferenceService"
COUNT_DOWN_LATCH_SERVICE = "hz:impl:countDownLatchService"
ID_GENERATOR_SERVICE = "hz:impl:idGeneratorService"
EXECUTOR_SERVICE = "hz:impl:executorService"
LOCK_SERVICE = "hz:impl:lockService"
LIST_SERVICE = "hz:impl:listService"
MULTI_MAP_SERVICE = "hz:impl:multiMapService"
MAP_SERVICE = "hz:impl:mapService"
RELIABLE_TOPIC_SERVICE = "hz:impl:reliableTopicService"
REPLICATED_MAP_SERVICE = "hz:impl:replicatedMapService"
RINGBUFFER_SERIVCE = "hz:impl:ringbufferService"
SEMAPHORE_SERVICE = "hz:impl:semaphoreService"
SET_SERVICE = "hz:impl:setService"
QUEUE_SERVICE = "hz:impl:queueService"
TOPIC_SERVICE = "hz:impl:topicService"
PN_COUNTER_SERVICE = "hz:impl:PNCounterService"
FLAKE_ID_GENERATOR_SERVICE = "hz:impl:flakeIdGeneratorService"

ID_GENERATOR_ATOMIC_LONG_PREFIX = "hz:atomic:idGenerator:"

_proxy_init = {
    ATOMIC_LONG_SERVICE: AtomicLong,
    ATOMIC_REFERENCE_SERVICE: AtomicReference,
    COUNT_DOWN_LATCH_SERVICE: CountDownLatch,
    ID_GENERATOR_SERVICE: IdGenerator,
    EXECUTOR_SERVICE: Executor,
    LIST_SERVICE: List,
    LOCK_SERVICE: Lock,
    MAP_SERVICE: create_map_proxy,
    MULTI_MAP_SERVICE: MultiMap,
    QUEUE_SERVICE: Queue,
    RELIABLE_TOPIC_SERVICE: ReliableTopic,
    REPLICATED_MAP_SERVICE: ReplicatedMap,
    RINGBUFFER_SERIVCE: Ringbuffer,
    SEMAPHORE_SERVICE: Semaphore,
    SET_SERVICE: Set,
    TOPIC_SERVICE: Topic,
    PN_COUNTER_SERVICE: PNCounter,
    FLAKE_ID_GENERATOR_SERVICE: FlakeIdGenerator
}


class ProxyManager(object):
    def __init__(self, client):
        self._client = client
        self._proxies = {}

    def get_or_create(self, service_name, name, **kwargs):
        ns = (service_name, name)
        if ns in self._proxies:
            return self._proxies[ns]

        proxy = self.create_proxy(service_name, name, **kwargs)
        self._proxies[ns] = proxy
        return proxy

    def create_proxy(self, service_name, name, **kwargs):
        message = client_create_proxy_codec.encode_request(name=name, service_name=service_name,
                                                           target=self._find_next_proxy_address())
        self._client.invoker.invoke_on_random_target(message).result()
        return _proxy_init[service_name](client=self._client, service_name=service_name, name=name, **kwargs)

    def destroy_proxy(self, service_name, name):
        ns = (service_name, name)
        try:
            self._proxies.pop(ns)
            message = client_destroy_proxy_codec.encode_request(name=name, service_name=service_name)
            self._client.invoker.invoke_on_random_target(message).result()
            return True
        except KeyError:
            return False

    def _find_next_proxy_address(self):
        # TODO: filter out lite members
        return self._client.load_balancer.next_address()



from uuid import uuid4
from hazelcast import future
from hazelcast.protocol.codec import executor_service_submit_to_address_codec, executor_service_shutdown_codec, \
    executor_service_is_shutdown_codec, executor_service_cancel_on_address_codec, \
    executor_service_cancel_on_partition_codec, executor_service_submit_to_partition_codec
from hazelcast.proxy.base import Proxy
from hazelcast.util import check_not_none


class Executor(Proxy):
    """
    An object that executes submitted executable tasks.
    """
    # TODO: cancellation
    def execute_on_key_owner(self, key, task):
        """
        Executes a task on the owner of the specified key.

        :param key: (object), the specified key.
        :param task: (Task), a task executed on the owner of the specified key.
        :return: (:class:`~hazelcast.future.Future`), future representing pending completion of the task.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        partition_id = self._client.partition_service.get_partition_id(key_data)

        uuid = self._get_uuid()
        return self._encode_invoke_on_partition(executor_service_submit_to_partition_codec, partition_id,
                                                uuid=uuid, callable=self._to_data(task),
                                                partition_id=partition_id)

    def execute_on_member(self, member, task):
        """
        Executes a task on the specified member.

        :param member: (Member), the specified member.
        :param task: (Task), the task executed on the specified member.
        :return: (:class:`~hazelcast.future.Future`), Future representing pending completion of the task.
        """
        uuid = self._get_uuid()
        address = member.address
        return self._execute_on_member(address, uuid, self._to_data(task))

    def execute_on_members(self, members, task):
        """
        Executes a task on each of the specified members.

        :param members: (Collection), the specified members.
        :param task: (Task), the task executed on the specified members.
        :return: (Map), :class:`~hazelcast.future.Future` tuples representing pending completion of the task on each member.
        """
        task_data = self._to_data(task)
        futures = []
        uuid = self._get_uuid()
        for member in members:
            f = self._execute_on_member(member.address, uuid, task_data)
            futures.append(f)
        return future.combine_futures(*futures)

    def execute_on_all_members(self, task):
        """
        Executes a task on all of the known cluster members.

        :param task: (Task), the task executed on the all of the members.
        :return: (Map), :class:`~hazelcast.future.Future` tuples representing pending completion of the task on each member.
        """
        return self.execute_on_members(self._client.cluster.get_member_list(), task)

    def is_shutdown(self):
        """
        Determines whether this executor has been shut down or not.

        :return: (bool), ``true`` if this executor has been shut down.
        """
        return self._encode_invoke(executor_service_is_shutdown_codec)

    def shutdown(self):
        """
        Initiates a shutdown process which works orderly. Tasks that were submitted before shutdown are executed but new
        task will not be accepted.
        """
        return self._encode_invoke(executor_service_shutdown_codec)

    def _execute_on_member(self, address, uuid, task_data):
        return self._encode_invoke_on_target(executor_service_submit_to_address_codec, address, uuid=uuid,
                                             callable=task_data, address=address)

    def _get_uuid(self):
        return str(uuid4())



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.codec.transactional_map_message_type import *

REQUEST_TYPE = TRANSACTIONALMAP_GET
RESPONSE_TYPE = 105
RETRYABLE = False


def calculate_size(name, txn_id, thread_id, key):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += calculate_size_str(txn_id)
    data_size += LONG_SIZE_IN_BYTES
    data_size += calculate_size_data(key)
    return data_size


def encode_request(name, txn_id, thread_id, key):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, txn_id, thread_id, key))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_str(txn_id)
    client_message.append_long(thread_id)
    client_message.append_data(key)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(response=None)
    if not client_message.read_bool():
        parameters['response'] = to_object(client_message.read_data())
    return parameters




SET_SIZE = 0x0601
SET_CONTAINS = 0x0602
SET_CONTAINSALL = 0x0603
SET_ADD = 0x0604
SET_REMOVE = 0x0605
SET_ADDALL = 0x0606
SET_COMPAREANDREMOVEALL = 0x0607
SET_COMPAREANDRETAINALL = 0x0608
SET_CLEAR = 0x0609
SET_GETALL = 0x060a
SET_ADDLISTENER = 0x060b
SET_REMOVELISTENER = 0x060c
SET_ISEMPTY = 0x060d



import logging
import os

from hazelcast.protocol.codec import client_statistics_codec
from hazelcast.util import calculate_version, current_time_in_millis, to_millis, to_nanos, current_time
from hazelcast.config import ClientProperties
from hazelcast.version import CLIENT_VERSION, CLIENT_TYPE
from hazelcast import six

try:
    import psutil
    PSUTIL_ENABLED = True
except ImportError:
    PSUTIL_ENABLED = False


class Statistics(object):

    _SINCE_VERSION_STRING = "3.9"
    _SINCE_VERSION = calculate_version(_SINCE_VERSION_STRING)

    _NEAR_CACHE_CATEGORY_PREFIX = "nc."
    _STAT_SEPARATOR = ","
    _KEY_VALUE_SEPARATOR = "="
    _EMPTY_STAT_VALUE = ""

    _DEFAULT_PROBE_VALUE = 0

    logger = logging.getLogger("HazelcastClient.Statistics")

    def __init__(self, client):
        self._client = client
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}
        self._enabled = client.properties.get_bool(ClientProperties.STATISTICS_ENABLED)
        self._cached_owner_address = None
        self._statistics_timer = None
        self._failed_gauges = set()

    def start(self):
        if not self._enabled:
            return

        period = self._client.properties.get_seconds(ClientProperties.STATISTICS_PERIOD_SECONDS)
        if period <= 0:
            default_period = self._client.properties.get_seconds_positive_or_default(
                ClientProperties.STATISTICS_PERIOD_SECONDS)

            self.logger.warning("Provided client statistics {} cannot be less than or equal to 0."
                                "You provided {} as the configuration. Client will use the default value "
                                "{} instead.".format(ClientProperties.STATISTICS_PERIOD_SECONDS.name,
                                                     period,
                                                     default_period), extra=self._logger_extras)
            period = default_period

        def _statistics_task():
            self._send_statistics()
            self._statistics_timer = self._client.reactor.add_timer(period, _statistics_task)

        self._statistics_timer = self._client.reactor.add_timer(period, _statistics_task)

        self.logger.info("Client statistics enabled with the period of {} seconds.".format(period),
                         extra=self._logger_extras)

    def shutdown(self):
        if self._statistics_timer:
            self._statistics_timer.cancel()

    def _send_statistics(self):
        owner_connection = self._get_owner_connection()
        if owner_connection is None:
            self.logger.debug("Cannot send client statistics to the server. No owner connection.",
                              extra=self._logger_extras)
            return

        stats = []
        self._fill_metrics(stats, owner_connection)
        self._add_near_cache_stats(stats)
        self._add_runtime_and_os_stats(stats)
        self._send_stats_to_owner("".join(stats), owner_connection)

    def _send_stats_to_owner(self, stats, owner_connection):
        request = client_statistics_codec.encode_request(stats)
        self._client.invoker.invoke_on_connection(request, owner_connection)

    def _add_runtime_and_os_stats(self, stats):
        os_and_runtime_stats = self._get_os_and_runtime_stats()
        for stat_name, stat_value in six.iteritems(os_and_runtime_stats):
            self._add_stat(stats, stat_name, stat_value)

    def _get_os_and_runtime_stats(self):
        psutil_stats = {}
        if PSUTIL_ENABLED:

            if self._can_collect_stat("os.totalPhysicalMemorySize") \
                    or self._can_collect_stat("os.freePhysicalMemorySize"):
                self._collect_physical_memory_info(psutil_stats, "os.totalPhysicalMemorySize",
                                                   "os.freePhysicalMemorySize")

            if self._can_collect_stat("os.totalSwapSpaceSize") or self._can_collect_stat("os.freeSwapSpaceSize"):
                self._collect_swap_memory_info(psutil_stats, "os.totalSwapSpaceSize", "os.freeSwapSpaceSize")

            if self._can_collect_stat("os.systemLoadAverage"):
                self._collect_load_average(psutil_stats, "os.systemLoadAverage")

            if self._can_collect_stat("runtime.availableProcessors"):
                self._collect_cpu_count(psutil_stats, "runtime.availableProcessors")

            process = psutil.Process()
            with process.oneshot():
                # With oneshot, process related information could be gathered
                # faster due to caching.

                if self._can_collect_stat("os.committedVirtualMemorySize") \
                        or self._can_collect_stat("runtime.usedMemory"):
                    self._collect_process_memory_info(psutil_stats, "os.committedVirtualMemorySize",
                                                      "runtime.usedMemory", process)

                if self._can_collect_stat("os.openFileDescriptorCount"):
                    self._collect_file_descriptor_count(psutil_stats, "os.openFileDescriptorCount", process)

                if self._can_collect_stat("os.maxFileDescriptorCount"):
                    self._collect_max_file_descriptor_count(psutil_stats, "os.maxFileDescriptorCount", process)

                if self._can_collect_stat("os.processCpuTime"):
                    self._collect_process_cpu_time(psutil_stats, "os.processCpuTime", process)

                if self._can_collect_stat("runtime.uptime"):
                    self._collect_process_uptime(psutil_stats, "runtime.uptime", process)

        return psutil_stats

    def _fill_metrics(self, stats, owner_connection):
        self._add_stat(stats, "lastStatisticsCollectionTime", current_time_in_millis())
        self._add_stat(stats, "enterprise", "false")
        self._add_stat(stats, "clientType", CLIENT_TYPE)
        self._add_stat(stats, "clientVersion", CLIENT_VERSION)
        self._add_stat(stats, "clusterConnectionTimestamp", to_millis(owner_connection.start_time_in_seconds))

        local_host, local_ip = owner_connection.socket.getsockname()
        local_address = str(local_host) + ":" + str(local_ip)
        self._add_stat(stats, "clientAddress", local_address)
        self._add_stat(stats, "clientName", self._client.name)

    def _add_near_cache_stats(self, stats):
        for near_cache in self._client.near_cache_manager.list_all_near_caches():
            near_cache_name_with_prefix = self._get_name_with_prefix(near_cache.name)
            near_cache_name_with_prefix.append(".")
            prefix = "".join(near_cache_name_with_prefix)

            near_cache_stats = near_cache.get_statistics()
            self._add_stat(stats, "creationTime", to_millis(near_cache_stats["creation_time"]), prefix)
            self._add_stat(stats, "evictions", near_cache_stats["evictions"], prefix)
            self._add_stat(stats, "hits", near_cache_stats["hits"], prefix)
            self._add_stat(stats, "misses", near_cache_stats["misses"], prefix)
            self._add_stat(stats, "ownedEntryCount", near_cache_stats["owned_entry_count"], prefix)
            self._add_stat(stats, "expirations", near_cache_stats["expirations"], prefix)
            self._add_stat(stats, "invalidations", near_cache_stats["invalidations"], prefix)
            self._add_stat(stats, "invalidationRequests", near_cache_stats["invalidation_requests"], prefix)
            self._add_stat(stats, "ownedEntryMemoryCost", near_cache_stats["owned_entry_memory_cost"], prefix)

    def _add_stat(self, stats, name, value, key_prefix=None):
        if len(stats) != 0:
            stats.append(Statistics._STAT_SEPARATOR)

        if key_prefix:
            stats.append(key_prefix)

        stats.append(name)
        stats.append(Statistics._KEY_VALUE_SEPARATOR)
        stats.append(str(value))

    def _add_empty_stat(self, stats, name, key_prefix=None):
        self._add_stat(stats, name, Statistics._EMPTY_STAT_VALUE, key_prefix)

    def _get_owner_connection(self):
        current_owner_address = self._client.cluster.owner_connection_address
        connection = self._client.connection_manager.get_connection(current_owner_address)

        if connection is None:
            return None

        server_version = connection.server_version
        if server_version < Statistics._SINCE_VERSION:
            # do not print too many logs if connected to an old version server
            if self._cached_owner_address and self._cached_owner_address != current_owner_address:
                self.logger.debug("Client statistics cannot be sent to server {} since,"
                                  "connected owner server version is less than the minimum supported server version ,"
                                  "{}.".format(current_owner_address, Statistics._SINCE_VERSION_STRING),
                                  extra=self._logger_extras)

            # cache the last connected server address for decreasing the log prints
            self._cached_owner_address = current_owner_address
            return None

        return connection

    def _get_name_with_prefix(self, name):
        return [Statistics._NEAR_CACHE_CATEGORY_PREFIX, self._escape_special_characters(name)]

    def _escape_special_characters(self, name):
        escaped_name = name.replace("\\", "\\\\").replace(",", "\\,").replace(".", "\\.").replace("=", "\\=")
        return escaped_name[1:] if name[0] == "/" else escaped_name

    def _can_collect_stat(self, name):
        return name not in self._failed_gauges

    def _safe_psutil_stat_collector(func):
        def safe_wrapper(self, psutil_stats, probe_name, *args):
            # Decorated function's signature must match with above
            try:
                stat = func(self, psutil_stats, probe_name, *args)
            except AttributeError as ae:
                self.logger.debug("Unable to register psutil method used for the probe {}. "
                                  "Cause: {}".format(probe_name, ae), extra=self._logger_extras)
                self._failed_gauges.add(probe_name)
                return
            except Exception as ex:
                self.logger.warning("Failed to access the probe {}. Cause: {}".format(probe_name, ex),
                                    extra=self._logger_extras)
                stat = self._DEFAULT_PROBE_VALUE

            psutil_stats[probe_name] = stat

        return safe_wrapper

    def _collect_physical_memory_info(self, psutil_stats, probe_total, probe_free):
        memory_info = psutil.virtual_memory()

        if self._can_collect_stat(probe_total):
            self._collect_total_physical_memory_size(psutil_stats, probe_total, memory_info)

        if self._can_collect_stat(probe_free):
            self._collect_free_physical_memory_size(psutil_stats, probe_free, memory_info)

    @_safe_psutil_stat_collector
    def _collect_total_physical_memory_size(self, psutil_stats, probe_name, memory_info):
        return memory_info.total

    @_safe_psutil_stat_collector
    def _collect_free_physical_memory_size(self, psutil_stats, probe_name, memory_info):
        return memory_info.available

    def _collect_swap_memory_info(self, psutil_stats, probe_total, probe_free):
        swap_info = psutil.swap_memory()

        if self._can_collect_stat(probe_total):
            self._collect_total_swap_space(psutil_stats, probe_total, swap_info)

        if self._can_collect_stat(probe_free):
            self._collect_free_swap_space(psutil_stats, probe_free, swap_info)

    @_safe_psutil_stat_collector
    def _collect_total_swap_space(self, psutil_stats, probe_name, swap_info):
        return swap_info.total

    @_safe_psutil_stat_collector
    def _collect_free_swap_space(self, psutil_stats, probe_name, swap_info):
        return swap_info.free

    @_safe_psutil_stat_collector
    def _collect_load_average(self, psutil_stats, probe_name):
        return os.getloadavg()[0]

    @_safe_psutil_stat_collector
    def _collect_cpu_count(self, psutil_stats, probe_name):
        return psutil.cpu_count()

    def _collect_process_memory_info(self, psutil_stats, probe_cvms, probe_used, process):
        p_memory_info = process.memory_info()

        if self._can_collect_stat(probe_cvms):
            self._collect_committed_virtual_memory_size(psutil_stats, probe_cvms, p_memory_info)

        if self._can_collect_stat(probe_used):
            self._collect_used_memory(psutil_stats, probe_used, p_memory_info)

    @_safe_psutil_stat_collector
    def _collect_committed_virtual_memory_size(self, psutil_stats, probe_name, p_memory_info):
        return p_memory_info.vms

    @_safe_psutil_stat_collector
    def _collect_used_memory(self, psutil_stats, probe_name, p_memory_info):
        return p_memory_info.rss

    @_safe_psutil_stat_collector
    def _collect_file_descriptor_count(self, psutil_stats, probe_name, process):
        try:
            return process.num_fds()
        except AttributeError:
            # Fallback for Windows
            return process.num_handles()

    @_safe_psutil_stat_collector
    def _collect_max_file_descriptor_count(self, psutil_stats, probe_name, process):
        return process.rlimit(psutil.RLIMIT_NOFILE)[1]

    @_safe_psutil_stat_collector
    def _collect_process_cpu_time(self, psutil_stats, probe_name, process):
        return to_nanos(sum(process.cpu_times()))

    @_safe_psutil_stat_collector
    def _collect_process_uptime(self, psutil_stats, probe_name, process):
        return to_millis(current_time() - process.create_time())





import math
from hazelcast.six.moves import range

def _fmix(h):
    h ^= h >> 16
    h = (h * 0x85ebca6b) & 0xFFFFFFFF
    h ^= h >> 13
    h = (h * 0xc2b2ae35) & 0xFFFFFFFF
    h ^= h >> 16
    return h


def murmur_hash3_x86_32(data, offset, size, seed=0x01000193):
    """
    murmur3 hash function to determine partition

    :param data: (byte array), input byte array
    :param offset: (long), offset.
    :param size: (long), byte length.
    :param seed: murmur hash seed hazelcast uses 0x01000193
    :return: (int32), calculated hash value.
    """
    key = bytearray(data[offset: offset + size])
    length = len(key)
    nblocks = int(length / 4)

    h1 = seed

    c1 = 0xcc9e2d51
    c2 = 0x1b873593

    # body
    for block_start in range(0, nblocks * 4, 4):
        # ??? big endian?
        k1 = key[block_start + 3] << 24 | \
             key[block_start + 2] << 16 | \
             key[block_start + 1] << 8 | \
             key[block_start + 0]

        k1 = c1 * k1 & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (c2 * k1) & 0xFFFFFFFF

        h1 ^= k1
        h1 = (h1 << 13 | h1 >> 19) & 0xFFFFFFFF  # inlined _ROTL32
        h1 = (h1 * 5 + 0xe6546b64) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 4
    k1 = 0
    tail_size = length & 3

    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size != 0:
        k1 = (k1 * c1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # _ROTL32
        k1 = (k1 * c2) & 0xFFFFFFFF
        h1 ^= k1

    result = _fmix(h1 ^ length)
    return -(result & 0x80000000) | (result & 0x7FFFFFFF)


def hash_to_index(hash, length):
    if hash == 0x80000000:
        return 0
    else:
        return int(abs(math.fmod(hash, length)))



import logging

from hazelcast.future import make_blocking
from hazelcast.partition import string_partition_strategy
from hazelcast.util import enum, thread_id
from hazelcast import six

MAX_SIZE = float('inf')


def default_response_handler(future, codec, to_object):
    response = future.result()
    if response:
        try:
            codec.decode_response
        except AttributeError:
            return
        decoded_response = codec.decode_response(response, to_object)
        try:
            return decoded_response['response']
        except AttributeError:
            pass


class Proxy(object):
    """
    Provides basic functionality for Hazelcast Proxies.
    """
    def __init__(self, client, service_name, name):
        self.service_name = service_name
        self.name = name
        self.partition_key = string_partition_strategy(self.name)
        self._client = client
        self.logger = logging.getLogger("HazelcastClient.%s(%s)" % (type(self).__name__, name))
        self._to_object = client.serialization_service.to_object
        self._to_data = client.serialization_service.to_data
        self._register_listener = client.listener.register_listener
        self._deregister_listener = client.listener.deregister_listener
        self._is_smart = client.listener.is_smart

    def destroy(self):
        """
        Destroys this proxy.

        :return: (bool), ``true`` if this proxy is deleted successfully, ``false`` otherwise.
        """
        self._on_destroy()
        return self._client.proxy.destroy_proxy(self.service_name, self.name)

    def _on_destroy(self):
        pass

    def __repr__(self):
        return '%s(name="%s")' % (type(self).__name__, self.name)

    def _encode_invoke(self, codec, response_handler=default_response_handler, **kwargs):
        request = codec.encode_request(name=self.name, **kwargs)
        return self._client.invoker.invoke_on_random_target(request).continue_with(response_handler, codec, self._to_object)

    def _encode_invoke_on_target(self, codec, _address, response_handler=default_response_handler, **kwargs):
        request = codec.encode_request(name=self.name, **kwargs)
        return self._client.invoker.invoke_on_target(request, _address).continue_with(response_handler, codec, self._to_object)

    def _encode_invoke_on_key(self, codec, key_data, invocation_timeout=None, **kwargs):
        partition_id = self._client.partition_service.get_partition_id(key_data)
        return self._encode_invoke_on_partition(codec, partition_id, invocation_timeout=invocation_timeout, **kwargs)

    def _encode_invoke_on_partition(self, codec, _partition_id, response_handler=default_response_handler,
                                    invocation_timeout=None, **kwargs):
        request = codec.encode_request(name=self.name, **kwargs)
        return self._client.invoker.invoke_on_partition(request, _partition_id, invocation_timeout).continue_with(response_handler,
                                                                                                                  codec, self._to_object)

    def blocking(self):
        """
        Returns a version of this proxy with only blocking method calls.
        :return: (Proxy), a version of this proxy with only blocking method calls.
        """
        return make_blocking(self)


class PartitionSpecificProxy(Proxy):
    """
    Provides basic functionality for Partition Specific Proxies.
    """
    def __init__(self, client, service_name, name):
        super(PartitionSpecificProxy, self).__init__(client, service_name, name)
        self._partition_id = self._client.partition_service.get_partition_id(self.partition_key)

    def _encode_invoke(self, codec, response_handler=default_response_handler, invocation_timeout=None, **kwargs):
        return super(PartitionSpecificProxy, self)._encode_invoke_on_partition(codec, self._partition_id,
                                                                               response_handler=response_handler,
                                                                               invocation_timeout=invocation_timeout, **kwargs)


class TransactionalProxy(object):
    """
    Provides an interface for all transactional distributed objects.
    """
    def __init__(self, name, transaction):
        self.name = name
        self.transaction = transaction
        self._to_object = transaction.client.serialization_service.to_object
        self._to_data = transaction.client.serialization_service.to_data

    def _encode_invoke(self, codec, response_handler=default_response_handler, **kwargs):
        request = codec.encode_request(name=self.name, txn_id=self.transaction.id, thread_id=thread_id(), **kwargs)
        return self.transaction.client.invoker.invoke_on_connection(request, self.transaction.connection).continue_with(
                response_handler, codec, self._to_object)

    def __repr__(self):
        return '%s(name="%s")' % (type(self).__name__, self.name)


ItemEventType = enum(added=1, removed=2)
EntryEventType = enum(added=1, removed=2, updated=4, evicted=8, evict_all=16, clear_all=32, merged=64, expired=128, invalidation=256)


class ItemEvent(object):
    """
    Map Item event.
    """
    def __init__(self, name, item_data, event_type, member, to_object):
        self.name = name
        self._item_data = item_data
        self.event_type = event_type
        self.member = member
        self._to_object = to_object

    @property
    def item(self):
        """The item related to the event."""
        return self._to_object(self._item_data)


class EntryEvent(object):
    """
    Map Entry event.
    """
    def __init__(self, to_object, key, old_value, value, merging_value, event_type, uuid,
                 number_of_affected_entries):
        self._key_data = key
        self._value_data = value
        self._old_value_data = old_value
        self._merging_value_data = merging_value
        self.event_type = event_type
        self.uuid = uuid
        self.number_of_affected_entries = number_of_affected_entries
        self._to_object = to_object

    @property
    def key(self):
        """The key of this entry event."""
        return self._to_object(self._key_data)

    @property
    def old_value(self):
        """The old value of the entry event."""
        return self._to_object(self._old_value_data)

    @property
    def value(self):
        """The value of the entry event."""
        return self._to_object(self._value_data)

    @property
    def merging_value(self):
        """The incoming merging value of the entry event."""
        return self._to_object(self._merging_value_data)

    def __repr__(self):
        return "EntryEvent(key=%s, old_value=%s, value=%s, merging_value=%s, event_type=%s, uuid=%s, " \
               "number_of_affected_entries=%s)" % (
                   self.key, self.old_value, self.value, self.merging_value, self.event_type, self.uuid,
                   self.number_of_affected_entries)


class TopicMessage(object):
    """
    Topic message.
    """
    def __init__(self, name, message_data, publish_time, member, to_object):
        self.name = name
        self._message_data = message_data
        self.publish_time = publish_time
        self.member = member
        self._to_object = to_object

    @property
    def message(self):
        """The message sent to Topic."""
        return self._to_object(self._message_data)


def get_entry_listener_flags(**kwargs):
    flags = 0
    for (key, value) in six.iteritems(kwargs):
        if value:
            flags |= getattr(EntryEventType, key)
    return flags



import json
import logging

from hazelcast.exception import HazelcastCertificationError
from hazelcast.util import _parse_address
from hazelcast.core import Address
from hazelcast.config import ClientProperty
from hazelcast.six.moves import http_client

try:
    import ssl
except ImportError:
    ssl = None


class HazelcastCloudAddressProvider(object):
    """
    Provides initial addresses for client to find and connect to a node.
    """

    logger = logging.getLogger("HazelcastClient.HazelcastCloudAddressProvider")

    def __init__(self, host, url, connection_timeout, logger_extras=None):
        self.cloud_discovery = HazelcastCloudDiscovery(host, url, connection_timeout)
        self._logger_extras = logger_extras

    def load_addresses(self):
        """
        Loads member addresses from Hazelcast.cloud endpoint.

        :return: (Sequence), The possible member addresses to connect to.
        """
        try:
            return list(self.cloud_discovery.discover_nodes().keys())
        except Exception as ex:
            self.logger.warning("Failed to load addresses from Hazelcast.cloud: {}".format(ex.args[0]),
                                extra=self._logger_extras)
        return []


class HazelcastCloudAddressTranslator(object):
    """
    Resolves private IP addresses of Hazelcast.cloud service.
    """

    logger = logging.getLogger("HazelcastClient.HazelcastCloudAddressTranslator")

    def __init__(self, host, url, connection_timeout, logger_extras=None):
        self.cloud_discovery = HazelcastCloudDiscovery(host, url, connection_timeout)
        self._private_to_public = dict()
        self._logger_extras = logger_extras

    def translate(self, address):
        """
        Translates the given address to another address specific to network or service.

        :param address: (:class:`~hazelcast.core.Address`), private address to be translated
        :return: (:class:`~hazelcast.core.Address`), new address if given address is known, otherwise returns null
        """
        if address is None:
            return None

        public_address = self._private_to_public.get(address)
        if public_address:
            return public_address

        self.refresh()

        return self._private_to_public.get(address)

    def refresh(self):
        """
        Refreshes the internal lookup table if necessary.
        """
        try:
            self._private_to_public = self.cloud_discovery.discover_nodes()
        except Exception as ex:
            self.logger.warning("Failed to load addresses from Hazelcast.cloud: {}".format(ex.args[0]),
                                extra=self._logger_extras)


class HazelcastCloudDiscovery(object):
    """
    Discovery service that discover nodes via Hazelcast.cloud
    https://coordinator.hazelcast.cloud/cluster/discovery?token=<TOKEN>
    """
    _CLOUD_URL_PATH = "/cluster/discovery?token="
    _PRIVATE_ADDRESS_PROPERTY = "private-address"
    _PUBLIC_ADDRESS_PROPERTY = "public-address"

    CLOUD_URL_BASE_PROPERTY = ClientProperty("hazelcast.client.cloud.url", "https://coordinator.hazelcast.cloud")
    """
    Internal client property to change base url of cloud discovery endpoint.
    Used for testing cloud discovery.
    """

    def __init__(self, host, url, connection_timeout):
        self._host = host
        self._url = url
        self._connection_timeout = connection_timeout
        # Default context operates only on TLSv1+, checks certificates,hostname and validity
        self._ctx = ssl.create_default_context()

    def discover_nodes(self):
        """
        Discovers nodes from Hazelcast.cloud.

        :return: (dict), Dictionary that maps private addresses to public addresses.
        """
        try:
            return self._call_service()
        except Exception as ex:
            raise ex

    def _call_service(self):
        try:
            https_connection = http_client.HTTPSConnection(host=self._host,
                                                           timeout=self._connection_timeout,
                                                           context=self._ctx)
            https_connection.request(method="GET", url=self._url, headers={"Accept-Charset": "UTF-8"})
            https_response = https_connection.getresponse()
        except ssl.SSLError as ex:
            raise HazelcastCertificationError(str(ex))
        self._check_error(https_response)
        return self._parse_response(https_response)

    def _check_error(self, https_response):
        response_code = https_response.status
        if response_code != 200:  # HTTP OK
            error_message = https_response.read().decode("utf-8")
            raise IOError(error_message)

    def _parse_response(self, https_response):
        json_value = json.loads(https_response.read().decode("utf-8"))
        private_to_public_addresses = dict()
        for value in json_value:
            private_address = value[self._PRIVATE_ADDRESS_PROPERTY]
            public_address = value[self._PUBLIC_ADDRESS_PROPERTY]

            public_addr = self._parse_address(public_address)
            private_addr = self._parse_address(private_address)
            if private_addr.port == -1:
                # If not explicitly given, set the port of the private address to port of the public address
                private_addr.port = public_addr.port
            private_to_public_addresses[private_addr] = public_addr

        return private_to_public_addresses

    def _parse_address(self, address):
        if ':' in address:
            host, port = address.split(':')
            return Address(host, int(port))
        return Address(address, -1)

    @staticmethod
    def get_host_and_url(properties, cloud_token):
        """
        Helper method to get host and url that can be used in HTTPSConnection.

        :param properties: Client config properties.
        :param cloud_token: Cloud discovery token.
        :return: Host and URL pair
        """
        host = properties.get(HazelcastCloudDiscovery.CLOUD_URL_BASE_PROPERTY.name,
                              HazelcastCloudDiscovery.CLOUD_URL_BASE_PROPERTY.default_value)
        host = host.replace("https://", "")
        host = host.replace("http://", "")
        return host, HazelcastCloudDiscovery._CLOUD_URL_PATH + cloud_token



import logging
import random
import threading
import time
import uuid

from hazelcast.exception import HazelcastError, AuthenticationError, TargetDisconnectedError
from hazelcast.lifecycle import LIFECYCLE_STATE_CONNECTED, LIFECYCLE_STATE_DISCONNECTED
from hazelcast.protocol.codec import client_add_membership_listener_codec, client_authentication_codec
from hazelcast.util import get_possible_addresses, get_provider_addresses, calculate_version
from hazelcast.version import CLIENT_TYPE, CLIENT_VERSION, SERIALIZATION_VERSION

# Membership Event Types
MEMBER_ADDED = 1
MEMBER_REMOVED = 2


class ClusterService(object):
    """
    Hazelcast cluster service. It provides access to the members in the cluster and the client can register for changes
    in the cluster members.

    All the methods on the Cluster are thread-safe.
    """
    logger = logging.getLogger("HazelcastClient.ClusterService")

    def __init__(self, config, client, address_providers):
        self._config = config
        self._client = client
        self._logger_extras = {"client_name": client.name, "group_name": config.group_config.name}
        self._members = {}
        self.owner_connection_address = None
        self.owner_uuid = None
        self.uuid = None
        self.listeners = {}

        for listener in config.membership_listeners:
            self.add_listener(*listener)

        self._address_providers = address_providers
        self._initial_list_fetched = threading.Event()
        self._client.connection_manager.add_listener(on_connection_closed=self._connection_closed)
        self._client.heartbeat.add_listener(on_heartbeat_stopped=self._heartbeat_stopped)

    def start(self):
        """
        Connects to cluster.
        """
        self._connect_to_cluster()

    def shutdown(self):
        pass

    def size(self):
        """
        Returns the size of the cluster.

        :return: (int), size of the cluster.
        """
        return len(self._members)

    def add_listener(self, member_added=None, member_removed=None, fire_for_existing=False):
        """
        Adds a membership listener to listen for membership updates, it will be notified when a member is added to
        cluster or removed from cluster. There is no check for duplicate registrations, so if you register the listener
        twice, it will get events twice.


        :param member_added: (Function), function to be called when a member is added to the cluster (optional).
        :param member_removed: (Function), function to be called when a member is removed to the cluster (optional).
        :param fire_for_existing: (bool), (optional).
        :return: (str), registration id of the listener which will be used for removing this listener.
        """
        registration_id = str(uuid.uuid4())
        self.listeners[registration_id] = (member_added, member_removed)

        if fire_for_existing:
            for member in self.get_member_list():
                member_added(member)

        return registration_id

    def remove_listener(self, registration_id):
        """
        Removes the specified membership listener.

        :param registration_id: (str), registration id of the listener to be deleted.
        :return: (bool), if the registration is removed, ``false`` otherwise.
        """
        try:
            self.listeners.pop(registration_id)
            return True
        except KeyError:
            return False

    @property
    def members(self):
        """
        Returns the members in the cluster.
        :return: (list), List of members.
        """
        return self.get_member_list()

    def _reconnect(self):
        try:
            self.logger.warning("Connection closed to owner node. Trying to reconnect.", extra=self._logger_extras)
            self._connect_to_cluster()
        except:
            self.logger.exception("Could not reconnect to cluster. Shutting down client.", extra=self._logger_extras)
            self._client.shutdown()

    def _connect_to_cluster(self):
        current_attempt = 1
        attempt_limit = self._config.network_config.connection_attempt_limit
        retry_delay = self._config.network_config.connection_attempt_period
        while current_attempt <= attempt_limit:
            provider_addresses = get_provider_addresses(self._address_providers)
            addresses = get_possible_addresses(provider_addresses, self.get_member_list())

            for address in addresses:
                try:
                    self.logger.info("Connecting to %s", address, extra=self._logger_extras)
                    self._connect_to_address(address)
                    return
                except:
                    self.logger.warning("Error connecting to %s ", address, exc_info=True, extra=self._logger_extras)

            if current_attempt >= attempt_limit:
                self.logger.warning(
                    "Unable to get alive cluster connection, attempt %d of %d",
                    current_attempt, attempt_limit, extra=self._logger_extras)
                break

            self.logger.warning(
                "Unable to get alive cluster connection, attempt %d of %d, trying again in %d seconds",
                current_attempt, attempt_limit, retry_delay, extra=self._logger_extras)
            current_attempt += 1
            time.sleep(retry_delay)

        error_msg = "Could not connect to any of %s after %d tries" % (addresses, attempt_limit)
        raise HazelcastError(error_msg)

    def _authenticate_manager(self, connection):
        request = client_authentication_codec.encode_request(
            username=self._config.group_config.name,
            password=self._config.group_config.password,
            uuid=self.uuid,
            owner_uuid=self.owner_uuid,
            is_owner_connection=True,
            client_type=CLIENT_TYPE,
            serialization_version=SERIALIZATION_VERSION,
            client_hazelcast_version=CLIENT_VERSION)

        def callback(f):
            parameters = client_authentication_codec.decode_response(f.result())
            if parameters["status"] != 0:  # TODO: handle other statuses
                raise AuthenticationError("Authentication failed.")
            connection.endpoint = parameters["address"]
            connection.is_owner = True
            self.owner_uuid = parameters["owner_uuid"]
            self.uuid = parameters["uuid"]
            connection.server_version_str = parameters.get("server_hazelcast_version", "")
            connection.server_version = calculate_version(connection.server_version_str)
            return connection

        return self._client.invoker.invoke_on_connection(request, connection).continue_with(callback)

    def _connect_to_address(self, address):
        f = self._client.connection_manager.get_or_connect(address, self._authenticate_manager)
        connection = f.result()
        if not connection.is_owner:
            self._authenticate_manager(connection).result()
        self.owner_connection_address = connection.endpoint
        self._init_membership_listener(connection)
        self._client.lifecycle.fire_lifecycle_event(LIFECYCLE_STATE_CONNECTED)

    def _init_membership_listener(self, connection):
        request = client_add_membership_listener_codec.encode_request(False)

        def handler(m):
            client_add_membership_listener_codec.handle(m, self._handle_member, self._handle_member_list)

        response = self._client.invoker.invoke_on_connection(request, connection, True, handler).result()
        registration_id = client_add_membership_listener_codec.decode_response(response)["response"]
        self.logger.debug("Registered membership listener with ID " + registration_id, extra=self._logger_extras)
        self._initial_list_fetched.wait()

    def _handle_member(self, member, event_type):
        self.logger.debug("Got member event: %s, %s", member, event_type, extra=self._logger_extras)
        if event_type == MEMBER_ADDED:
            self._member_added(member)
        elif event_type == MEMBER_REMOVED:
            self._member_removed(member)

        self._log_member_list()
        self._client.partition_service.refresh()

    def _handle_member_list(self, members):
        self.logger.debug("Got initial member list: %s", members, extra=self._logger_extras)

        for m in self.get_member_list():
            try:
                members.remove(m)
            except ValueError:
                self._member_removed(m)
        for m in members:
            self._member_added(m)

        self._log_member_list()
        self._client.partition_service.refresh()
        self._initial_list_fetched.set()

    def _member_added(self, member):
        self._members[member.address] = member
        for added, _ in list(self.listeners.values()):
            if added:
                try:
                    added(member)
                except:
                    self.logger.exception("Exception in membership listener", extra=self._logger_extras)

    def _member_removed(self, member):
        self._members.pop(member.address, None)
        self._client.connection_manager.close_connection(member.address, TargetDisconnectedError(
            "%s is no longer a member of the cluster" % member))
        for _, removed in list(self.listeners.values()):
            if removed:
                try:
                    removed(member)
                except:
                    self.logger.exception("Exception in membership listener", extra=self._logger_extras)

    def _log_member_list(self):
        self.logger.info("New member list:\n\nMembers [%d] {\n%s\n}\n", self.size(),
                         "\n".join(["\t" + str(x) for x in self.get_member_list()]), extra=self._logger_extras)

    def _connection_closed(self, connection, _):
        if connection.endpoint and connection.endpoint == self.owner_connection_address \
                and self._client.lifecycle.is_live:
            self._client.lifecycle.fire_lifecycle_event(LIFECYCLE_STATE_DISCONNECTED)
            self.owner_connection_address = None

            # try to reconnect, on new thread
            reconnect_thread = threading.Thread(target=self._reconnect,
                                                name="hazelcast-cluster-reconnect-{:.4}".format(str(uuid.uuid4())))
            reconnect_thread.daemon = True
            reconnect_thread.start()

    def _heartbeat_stopped(self, connection):
        if connection.endpoint == self.owner_connection_address:
            self._client.connection_manager.close_connection(connection.endpoint, TargetDisconnectedError(
                "%s stopped heart beating." % connection))

    def get_member_by_uuid(self, member_uuid):
        """
        Returns the member with specified member uuid.

        :param member_uuid: (int), uuid of the desired member.
        :return: (:class:`~hazelcast.core.Member`), the corresponding member.
        """
        for member in self.get_member_list():
            if member.uuid == member_uuid:
                return member

    def get_member_by_address(self, address):
        """
        Returns the member with the specified address if it is in the
        cluster, None otherwise.

        :param address: (:class:`~hazelcast.core.Address`), address of the desired member.
        :return: (:class:`~hazelcast.core.Member`), the corresponding member.
        """
        return self._members.get(address, None)

    def get_members(self, selector):
        """
        Returns the members that satisfy the given selector.

        :param selector: (:class:`~hazelcast.core.MemberSelector`), Selector to be applied to the members.
        :return: (List), List of members.
        """
        members = []
        for member in self.get_member_list():
            if selector.select(member):
                members.append(member)

        return members

    def get_member_list(self):
        """
        Returns all the members as a list.

        :return: (List), List of members.
        """
        return list(self._members.values())


class RandomLoadBalancer(object):
    """
    RandomLoadBalancer make the Client send operations randomly on members not to increase the load on a specific
    member.
    """

    def __init__(self, cluster):
        self._cluster = cluster

    def next_address(self):
        try:
            return random.choice(self._cluster.get_member_list()).address
        except IndexError:
            return None


class VectorClock(object):
    """
    Vector clock consisting of distinct replica logical clocks.

    See https://en.wikipedia.org/wiki/Vector_clock
    The vector clock may be read from different thread but concurrent
    updates must be synchronized externally. There is no guarantee for
    concurrent updates.
    """

    def __init__(self):
        self._replica_timestamps = {}

    def is_after(self, other):
        """
        Returns true if this vector clock is causally strictly after the
        provided vector clock. This means that it the provided clock is neither
        equal to, greater than or concurrent to this vector clock.

        :param other: (:class:`~hazelcast.cluster.VectorClock`), Vector clock to be compared
        :return: (bool), True if this vector clock is strictly after the other vector clock, False otherwise
        """
        any_timestamp_greater = False
        for replica_id, other_timestamp in other.entry_set():
            local_timestamp = self._replica_timestamps.get(replica_id)

            if local_timestamp is None or local_timestamp < other_timestamp:
                return False
            elif local_timestamp > other_timestamp:
                any_timestamp_greater = True

        # there is at least one local timestamp greater or local vector clock has additional timestamps
        return any_timestamp_greater or other.size() < self.size()

    def set_replica_timestamp(self, replica_id, timestamp):
        """
        Sets the logical timestamp for the given replica ID.

        :param replica_id: (str), Replica ID.
        :param timestamp: (int), Timestamp for the given replica ID.
        """
        self._replica_timestamps[replica_id] = timestamp

    def entry_set(self):
        """
        Returns the entry set of the replica timestamps in a format
        of list of tuples. Each tuple contains the replica ID and the
        timestamp associated with it.

        :return: (list), List of tuples.
        """
        return list(self._replica_timestamps.items())

    def size(self):
        """
        Returns the number of timestamps that are in the
        replica timestamps dictionary.

        :return: (int), Number of timestamps in the replica timestamps.
        """
        return len(self._replica_timestamps)



import itertools

from hazelcast.future import combine_futures, ImmediateFuture
from hazelcast.near_cache import NearCache
from hazelcast.protocol.codec import map_add_entry_listener_codec, map_add_entry_listener_to_key_codec, \
    map_add_entry_listener_with_predicate_codec, map_add_entry_listener_to_key_with_predicate_codec, \
    map_add_index_codec, map_clear_codec, map_contains_key_codec, map_contains_value_codec, map_delete_codec, \
    map_entry_set_codec, map_entries_with_predicate_codec, map_evict_codec, map_evict_all_codec, map_flush_codec, \
    map_force_unlock_codec, map_get_codec, map_get_all_codec, map_get_entry_view_codec, map_is_empty_codec, \
    map_is_locked_codec, map_key_set_codec, map_key_set_with_predicate_codec, map_load_all_codec, \
    map_load_given_keys_codec, map_lock_codec, map_put_codec, map_put_all_codec, map_put_if_absent_codec, \
    map_put_transient_codec, map_size_codec, map_remove_codec, map_remove_if_same_codec, \
    map_remove_entry_listener_codec, map_replace_codec, map_replace_if_same_codec, map_set_codec, map_try_lock_codec, \
    map_try_put_codec, map_try_remove_codec, map_unlock_codec, map_values_codec, map_values_with_predicate_codec, \
    map_add_interceptor_codec, map_execute_on_all_keys_codec, map_execute_on_key_codec, map_execute_on_keys_codec, \
    map_execute_with_predicate_codec, map_add_near_cache_entry_listener_codec
from hazelcast.proxy.base import Proxy, EntryEvent, EntryEventType, get_entry_listener_flags, MAX_SIZE
from hazelcast.util import check_not_none, thread_id, to_millis
from hazelcast import six


class Map(Proxy):
    """
    Hazelcast Map client proxy to access the map on the cluster.

    Concurrent, distributed, observable and queryable map.

    This map can work both async(non-blocking) or sync(blocking). Blocking calls return the value of the call and block
    the execution until return value is calculated. However, async calls return :class:`~hazelcast.future.Future` and do not block execution.
    Result of the :class:`~hazelcast.future.Future` can be used whenever ready. A :class:`~hazelcast.future.Future`'s result can be obtained with blocking the execution by
    calling :func:`Future.result() <hazelcast.future.Future.result>`.

        Example of sync(blocking) usage:
            >>> #client is initialized Hazelcast Client.
            >>> my_map = client.get_map("map").blocking()# returns sync map, all map functions are blocking
            >>> print("map.put", my_map.put("key", "value"))
            >>> print("map.contains_key", my_map.contains_key("key"))
            >>> print("map.get", my_map.get("key"))
            >>> print("map.size", my_map.size())

        Example of async(non-blocking) usage:
            >>> #client is initialized Hazelcast Client.
            >>> my_map = client.get_map("map")
            >>> def put_callback(f):
            >>>     print("map.put", f.result())
            >>> my_map.put("key", "async_val").add_done_callback(put_callback)
            >>>
            >>> print("map.size", my_map.size().result())
            >>>
            >>> def contains_key_callback(f):
            >>>     print("map.contains_key", f.result())
            >>> my_map.contains_key("key").add_done_callback(contains_key_callback)

    This class does not allow ``None`` to be used as a key or value.
    """
    def __init__(self, client, service_name, name):
        super(Map, self).__init__(client, service_name, name)
        self.reference_id_generator = self._client.lock_reference_id_generator

    def add_entry_listener(self, include_value=False, key=None, predicate=None, added_func=None, removed_func=None, updated_func=None,
                           evicted_func=None, evict_all_func=None, clear_all_func=None, merged_func=None, expired_func=None):
        """
        Adds a continuous entry listener for this map. Listener will get notified for map events filtered with given
        parameters.

        :param include_value: (bool), whether received event should include the value or not (optional).
        :param key: (object), key for filtering the events (optional).
        :param predicate: (Predicate), predicate for filtering the events (optional).
        :param added_func: Function to be called when an entry is added to map (optional).
        :param removed_func: Function to be called when an entry is removed from map (optional).
        :param updated_func: Function to be called when an entry is updated (optional).
        :param evicted_func: Function to be called when an entry is evicted from map (optional).
        :param evict_all_func: Function to be called when entries are evicted from map (optional).
        :param clear_all_func: Function to be called when entries are cleared from map (optional).
        :param merged_func: Function to be called when WAN replicated entry is merged_func (optional).
        :param expired_func: Function to be called when an entry's live time is expired (optional).
        :return: (str), a registration id which is used as a key to remove the listener.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        flags = get_entry_listener_flags(added=added_func, removed=removed_func, updated=updated_func,
                                         evicted=evicted_func, evict_all=evict_all_func, clear_all=clear_all_func,
                                         merged=merged_func, expired=expired_func)

        if key and predicate:
            key_data = self._to_data(key)
            predicate_data = self._to_data(predicate)
            request = map_add_entry_listener_to_key_with_predicate_codec.encode_request(
                self.name, key_data, predicate_data, include_value, flags, self._is_smart)
        elif key and not predicate:
            key_data = self._to_data(key)
            request = map_add_entry_listener_to_key_codec.encode_request(
                self.name, key_data, include_value, flags, self._is_smart)
        elif not key and predicate:
            predicate = self._to_data(predicate)
            request = map_add_entry_listener_with_predicate_codec.encode_request(
                self.name, predicate, include_value, flags, self._is_smart)
        else:
            request = map_add_entry_listener_codec.encode_request(self.name, include_value, flags, self._is_smart)

        def handle_event_entry(**_kwargs):
            event = EntryEvent(self._to_object, **_kwargs)
            if event.event_type == EntryEventType.added:
                added_func(event)
            elif event.event_type == EntryEventType.removed:
                removed_func(event)
            elif event.event_type == EntryEventType.updated:
                updated_func(event)
            elif event.event_type == EntryEventType.evicted:
                evicted_func(event)
            elif event.event_type == EntryEventType.evict_all:
                evict_all_func(event)
            elif event.event_type == EntryEventType.clear_all:
                clear_all_func(event)
            elif event.event_type == EntryEventType.merged:
                merged_func(event)
            elif event.event_type == EntryEventType.expired:
                expired_func(event)

        return self._register_listener(request, lambda r: map_add_entry_listener_codec.decode_response(r)['response'],
                                       lambda reg_id: map_remove_entry_listener_codec.encode_request(self.name, reg_id),
                                       lambda m: map_add_entry_listener_codec.handle(m, handle_event_entry))

    def add_index(self, attribute, ordered=False):
        """
        Adds an index to this map for the specified entries so that queries can run faster.

        Example:
            Let's say your map values are Employee objects.
                >>> class Employee(IdentifiedDataSerializable):
                >>>     active = false
                >>>     age = None
                >>>     name = None
                >>>     #other fields
                >>>
                >>>     #methods

            If you query your values mostly based on age and active fields, you should consider indexing these.
                >>> map = self.client.get_map("employees")
                >>> map.add_index("age" , true) #ordered, since we have ranged queries for this field
                >>> map.add_index("active", false) #not ordered, because boolean field cannot have range


        :param attribute: (str), index attribute of the value.
        :param ordered: (bool), for ordering the index or not (optional).
        """
        return self._encode_invoke(map_add_index_codec, attribute=attribute, ordered=ordered)

    def add_interceptor(self, interceptor):
        """
        Adds an interceptor for this map. Added interceptor will intercept operations and execute user defined methods.

        :param interceptor: (object), interceptor for the map which includes user defined methods.
        :return: (str),id of registered interceptor.
        """
        return self._encode_invoke(map_add_interceptor_codec, interceptor=self._to_data(interceptor))

    def clear(self):
        """
        Clears the map.

        The MAP_CLEARED event is fired for any registered listeners.
        """
        return self._encode_invoke(map_clear_codec)

    def contains_key(self, key):
        """
        Determines whether this map contains an entry with the key.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :return: (bool), ``true`` if this map contains an entry for the specified key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._contains_key_internal(key_data)

    def contains_value(self, value):
        """
        Determines whether this map contains one or more keys for the specified value.

        :param value: (object), the specified value.
        :return: (bool), ``true`` if this map contains an entry for the specified value.
        """
        check_not_none(value, "value can't be None")
        value_data = self._to_data(value)
        return self._encode_invoke(map_contains_value_codec, value=value_data)

    def delete(self, key):
        """
        Removes the mapping for a key from this map if it is present (optional operation).

        Unlike remove(object), this operation does not return the removed value, which avoids the serialization cost of
        the returned value. If the removed value will not be used, a delete operation is preferred over a remove
        operation for better performance.

        The map will not contain a mapping for the specified key once the call returns.

        **Warning:
        This method breaks the contract of EntryListener. When an entry is removed by delete(), it fires an EntryEvent
        with a** ``None`` **oldValue. Also, a listener with predicates will have** ``None`` **values, so only the keys can be queried
        via predicates.**

        :param key: (object), key of the mapping to be deleted.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._delete_internal(key_data)

    def entry_set(self, predicate=None):
        """
        Returns a list clone of the mappings contained in this map.

        **Warning:
        The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.**

        :param predicate: (Predicate), predicate for the map to filter entries (optional).
        :return: (Sequence), the list of key-value tuples in the map.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        if predicate:
            predicate_data = self._to_data(predicate)
            return self._encode_invoke(map_entries_with_predicate_codec, predicate=predicate_data)
        else:
            return self._encode_invoke(map_entry_set_codec)

    def evict(self, key):
        """
        Evicts the specified key from this map.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key to evict.
        :return: (bool), ``true`` if the key is evicted, ``false`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._evict_internal(key_data)

    def evict_all(self):
        """
        Evicts all keys from this map except the locked ones.

        The EVICT_ALL event is fired for any registered listeners.
        """
        return self._encode_invoke(map_evict_all_codec)

    def execute_on_entries(self, entry_processor, predicate=None):
        """
        Applies the user defined EntryProcessor to all the entries in the map or entries in the map which satisfies
        the predicate if provided. Returns the results mapped by each key in the map.

        :param entry_processor: (object), A stateful serializable object which represents the EntryProcessor defined on
            server side.
            This object must have a serializable EntryProcessor counter part registered on server side with the actual
            ``org.hazelcast.map.EntryProcessor`` implementation.
        :param predicate: (Predicate), predicate for filtering the entries (optional).
        :return: (Sequence), list of map entries which includes the keys and the results of the entry process.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        if predicate:
            return self._encode_invoke(map_execute_with_predicate_codec, entry_processor=self._to_data(entry_processor),
                                       predicate=self._to_data(predicate))
        return self._encode_invoke(map_execute_on_all_keys_codec, entry_processor=self._to_data(entry_processor))

    def execute_on_key(self, key, entry_processor):
        """
        Applies the user defined EntryProcessor to the entry mapped by the key. Returns the object which is the result
        of EntryProcessor's process method.

        :param key: (object), specified key for the entry to be processed.
        :param entry_processor: (object), A stateful serializable object which represents the EntryProcessor defined on
            server side.
            This object must have a serializable EntryProcessor counter part registered on server side with the actual
            ``org.hazelcast.map.EntryProcessor`` implementation.
        :return: (object), result of entry process.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._execute_on_key_internal(key_data, entry_processor)

    def execute_on_keys(self, keys, entry_processor):
        """
        Applies the user defined EntryProcessor to the entries mapped by the collection of keys. Returns the results
        mapped by each key in the collection.

        :param keys: (Collection), collection of the keys for the entries to be processed.
        :param entry_processor: (object), A stateful serializable object which represents the EntryProcessor defined on
            server side.
            This object must have a serializable EntryProcessor counter part registered on server side with the actual
            ``org.hazelcast.map.EntryProcessor`` implementation.
        :return: (Sequence), list of map entries which includes the keys and the results of the entry process.
        """
        key_list = []
        for key in keys:
            check_not_none(key, "key can't be None")
            key_list.append(self._to_data(key))

        if len(keys) == 0:
            return ImmediateFuture([])

        return self._encode_invoke(map_execute_on_keys_codec, entry_processor=self._to_data(entry_processor),
                                   keys=key_list)

    def flush(self):
        """
        Flushes all the local dirty entries.
        """
        return self._encode_invoke(map_flush_codec)

    def force_unlock(self, key):
        """
        Releases the lock for the specified key regardless of the lock owner. It always successfully unlocks the key,
        never blocks, and returns immediately.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the key to lock.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(map_force_unlock_codec, key_data, key=key_data,
                                          reference_id=self.reference_id_generator.get_and_increment())

    def get(self, key):
        """
        Returns the value for the specified key, or ``None`` if this map does not contain this key.

        **Warning:
        This method returns a clone of original value, modifying the returned value does not change the actual value in
        the map. One should put modified value back to make changes visible to all nodes.**
            >>> value = map.get(key)
            >>> value.update_some_property()
            >>> map.put(key,value)

        **Warning 2: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :return: (object), the value for the specified key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._get_internal(key_data)

    def get_all(self, keys):
        """
        Returns the entries for the given keys.

        **Warning:
        The returned map is NOT backed by the original map, so changes to the original map are NOT reflected in the
        returned map, and vice-versa.**

        **Warning 2: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param keys: (Collection), keys to get.
        :return: (dict), dictionary of map entries.
        """
        check_not_none(keys, "keys can't be None")
        if not keys:
            return ImmediateFuture({})

        partition_service = self._client.partition_service
        partition_to_keys = {}

        for key in keys:
            check_not_none(key, "key can't be None")
            key_data = self._to_data(key)
            partition_id = partition_service.get_partition_id(key_data)
            try:
                partition_to_keys[partition_id][key] = key_data
            except KeyError:
                partition_to_keys[partition_id] = {key: key_data}

        return self._get_all_internal(partition_to_keys)

    def get_entry_view(self, key):
        """
        Returns the EntryView for the specified key.

        **Warning:
        This method returns a clone of original mapping, modifying the returned value does not change the actual value
        in the map. One should put modified value back to make changes visible to all nodes.**

        **Warning 2: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the key of the entry.
        :return: (EntryView), EntryView of the specified key.

        .. seealso:: :class:`~hazelcast.core.EntryView` for more info about EntryView.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(map_get_entry_view_codec, key_data, key=key_data, thread_id=thread_id())

    def is_empty(self):
        """
        Returns ``true`` if this map contains no key-value mappings.

        :return: (bool), ``true`` if this map contains no key-value mappings.
        """
        return self._encode_invoke(map_is_empty_codec)

    def is_locked(self, key):
        """
        Checks the lock for the specified key. If the lock is acquired, it returns ``true``. Otherwise, it returns ``false``.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the key that is checked for lock
        :return: (bool), ``true`` if lock is acquired, ``false`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(map_is_locked_codec, key_data, key=key_data)

    def key_set(self, predicate=None):
        """
        Returns a List clone of the keys contained in this map or the keys of the entries filtered with the predicate if
        provided.

        **Warning:
        The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.**

        :param predicate: (Predicate), predicate to filter the entries (optional).
        :return: (Sequence), a list of the clone of the keys.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        if predicate:
            predicate_data = self._to_data(predicate)
            return self._encode_invoke(map_key_set_with_predicate_codec, predicate=predicate_data)
        else:
            return self._encode_invoke(map_key_set_codec)

    def load_all(self, keys=None, replace_existing_values=True):
        """
        Loads all keys from the store at server side or loads the given keys if provided.

        :param keys: (Collection), keys of the entry values to load (optional).
        :param replace_existing_values: (bool), whether the existing values will be replaced or not with those loaded
        from the server side MapLoader (optional).
        """
        if keys:
            key_data_list = list(map(self._to_data, keys))
            return self._load_all_internal(key_data_list, replace_existing_values)
        else:
            return self._encode_invoke(map_load_all_codec, replace_existing_values=replace_existing_values)

    def lock(self, key, ttl=-1):
        """
        Acquires the lock for the specified key infinitely or for the specified lease time if provided.

        If the lock is not available, the current thread becomes disabled for thread scheduling purposes and lies
        dormant until the lock has been acquired.

        You get a lock whether the value is present in the map or not. Other threads (possibly on other systems) would
        block on their invoke of lock() until the non-existent key is unlocked. If the lock holder introduces the key to
        the map, the put() operation is not blocked. If a thread not holding a lock on the non-existent key tries to
        introduce the key while a lock exists on the non-existent key, the put() operation blocks until it is unlocked.

        Scope of the lock is this map only. Acquired lock is only for the key in this map.

        Locks are re-entrant; so, if the key is locked N times, it should be unlocked N times before another thread can
        acquire it.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the key to lock.
        :param ttl: (int), time in seconds to wait before releasing the lock (optional).
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(map_lock_codec, key_data, invocation_timeout=MAX_SIZE, key=key_data,
                                          thread_id=thread_id(), ttl=to_millis(ttl),
                                          reference_id=self.reference_id_generator.get_and_increment())

    def put(self, key, value, ttl=-1):
        """
        Associates the specified value with the specified key in this map. If the map previously contained a mapping for
        the key, the old value is replaced by the specified value. If ttl is provided, entry will expire and get evicted
        after the ttl.

        **Warning:
        This method returns a clone of the previous value, not the original (identically equal) value previously put
        into the map.**

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :param value: (object), the value to associate with the key.
        :param ttl: (int), maximum time in seconds for this entry to stay, if not provided, the value configured on
            server side configuration will be used(optional).
        :return: (object), previous value associated with key or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_internal(key_data, value_data, ttl)

    def put_all(self, map):
        """
        Copies all of the mappings from the specified map to this map. No atomicity guarantees are
        given. In the case of a failure, some of the key-value tuples may get written, while others are not.

        :param map: (dict), map which includes mappings to be stored in this map.
        """
        check_not_none(map, "map can't be None")
        if not map:
            return ImmediateFuture(None)

        partition_service = self._client.partition_service
        partition_map = {}

        for key, value in six.iteritems(map):
            check_not_none(key, "key can't be None")
            check_not_none(value, "value can't be None")
            entry = (self._to_data(key), self._to_data(value))
            partition_id = partition_service.get_partition_id(entry[0])
            try:
                partition_map[partition_id].append(entry)
            except KeyError:
                partition_map[partition_id] = [entry]

        futures = []
        for partition_id, entry_list in six.iteritems(partition_map):
            future = self._encode_invoke_on_partition(map_put_all_codec, partition_id, entries=dict(entry_list))
            futures.append(future)

        return combine_futures(*futures)

    def put_if_absent(self, key, value, ttl=-1):
        """
        Associates the specified key with the given value if it is not already associated. If ttl is provided, entry
        will expire and get evicted after the ttl.

        This is equivalent to:
            >>> if not map.contains_key(key):
            >>>     return map.put(key,value)
            >>> else:
            >>>     return map.get(key)
        except that the action is performed atomically.

        **Warning:
        This method returns a clone of the previous value, not the original (identically equal) value previously put
        into the map.**

        **Warning 2: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key of the entry.
        :param value: (object), value of the entry.
        :param ttl:  (int), maximum time in seconds for this entry to stay in the map, if not provided, the value
            configured on server side configuration will be used (optional).
        :return: (object), old value of the entry.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_if_absent_internal(key_data, value_data, ttl)

    def put_transient(self, key, value, ttl=-1):
        """
        Same as put(TKey, TValue, Ttl), but MapStore defined at the server side will not be called.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key of the entry.
        :param value: (object), value of the entry.
        :param ttl: (int), maximum time in seconds for this entry to stay in the map, if not provided, the value
            configured on server side configuration will be used (optional).
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_transient_internal(key_data, value_data, ttl)

    def remove(self, key):
        """
        Removes the mapping for a key from this map if it is present. The map will not contain a mapping for the
        specified key once the call returns.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key of the mapping to be deleted.
        :return: (object), the previous value associated with key, or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._remove_internal(key_data)

    def remove_if_same(self, key, value):
        """
        Removes the entry for a key only if it is currently mapped to a given value.

        This is equivalent to:
            >>> if map.contains_key(key) and map.get(key).equals(value):
            >>>     map.remove(key)
            >>>     return true
            >>> else:
            >>>     return false
        except that the action is performed atomically.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :param value: (object), remove the key if it has this value.
        :return: (bool), ``true`` if the value was removed.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._remove_if_same_internal_(key_data, value_data)

    def remove_entry_listener(self, registration_id):
        """
        Removes the specified entry listener. Returns silently if there is no such listener added before.

        :param registration_id: (str), id of registered listener.
        :return: (bool), ``true`` if registration is removed, ``false`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def replace(self, key, value):
        """
        Replaces the entry for a key only if it is currently mapped to some value.

        This is equivalent to:
            >>> if map.contains_key(key):
            >>>     return map.put(key,value)
            >>> else:
            >>>     return None
        except that the action is performed atomically.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        **Warning 2:
        This method returns a clone of the previous value, not the original (identically equal) value previously put
        into the map.**

        :param key: (object), the specified key.
        :param value: (object), the value to replace the previous value.
        :return: (object), previous value associated with key, or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)

        return self._replace_internal(key_data, value_data)

    def replace_if_same(self, key, old_value, new_value):
        """
        Replaces the entry for a key only if it is currently mapped to a given value.

        This is equivalent to:
            >>> if map.contains_key(key) and map.get(key).equals(old_value):
            >>>     map.put(key, new_value)
            >>>     return true
            >>> else:
            >>>     return false
        except that the action is performed atomically.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :param old_value: (object), replace the key value if it is the old value.
        :param new_value: (object), the new value to replace the old value.
        :return: (bool), ``true`` if the value was replaced.
        """
        check_not_none(key, "key can't be None")
        check_not_none(old_value, "old_value can't be None")
        check_not_none(new_value, "new_value can't be None")

        key_data = self._to_data(key)
        old_value_data = self._to_data(old_value)
        new_value_data = self._to_data(new_value)

        return self._replace_if_same_internal(key_data, old_value_data, new_value_data)

    def set(self, key, value, ttl=-1):
        """
        Puts an entry into this map. Similar to the put operation except that set doesn't return the old value, which is
        more efficient. If ttl is provided, entry will expire and get evicted after the ttl.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key of the entry.
        :param value: (object), value of the entry.
        :param ttl: (int), maximum time in seconds for this entry to stay in the map, 0 means infinite. If ttl is not
            provided, the value configured on server side configuration will be used (optional).
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._set_internal(key_data, value_data, ttl)

    def size(self):
        """
        Returns the number of entries in this map.

        :return: (int), number of entries in this map.
        """
        return self._encode_invoke(map_size_codec)

    def try_lock(self, key, ttl=-1, timeout=0):
        """
        Tries to acquire the lock for the specified key. When the lock is not available,

            * If timeout is not provided, the current thread doesn't wait and returns ``false`` immediately.
            * If a timeout is provided, the current thread becomes disabled for thread scheduling purposes and lies
              dormant until one of the followings happens:
                * the lock is acquired by the current thread, or
                * the specified waiting time elapses.

        If ttl is provided, lock will be released after this time elapses.

        :param key: (object), key to lock in this map.
        :param ttl: (int), time in seconds to wait before releasing the lock (optional).
        :param timeout: (int), maximum time in seconds to wait for the lock (optional).
        :return: (bool), ``true`` if the lock was acquired and otherwise, ``false``.
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)

        return self._encode_invoke_on_key(map_try_lock_codec, key_data, invocation_timeout=MAX_SIZE, key=key_data,
                                          thread_id=thread_id(), lease=to_millis(ttl), timeout=to_millis(timeout),
                                          reference_id=self.reference_id_generator.get_and_increment())

    def try_put(self, key, value, timeout=0):
        """
        Tries to put the given key and value into this map and returns immediately if timeout is not provided. If
        timeout is provided, operation waits until it is completed or timeout is reached.

        :param key: (object), key of the entry.
        :param value: (object), value of the entry.
        :param timeout: (int), operation timeout in seconds (optional).
        :return: (bool), ``true`` if the put is successful, ``false`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)

        return self._try_put_internal(key_data, value_data, timeout)

    def try_remove(self, key, timeout=0):
        """
        Tries to remove the given key from this map and returns immediately if timeout is not provided. If
        timeout is provided, operation waits until it is completed or timeout is reached.

        :param key: (object), key of the entry to be deleted.
        :param timeout: (int), operation timeout in seconds (optional).
        :return: (bool), ``true`` if the remove is successful, ``false`` otherwise.
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)
        return self._try_remove_internal(key_data, timeout)

    def unlock(self, key):
        """
        Releases the lock for the specified key. It never blocks and returns immediately. If the current thread is the
        holder of this lock, then the hold count is decremented. If the hold count is zero, then the lock is released.

        :param key: (object), the key to lock.
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)

        return self._encode_invoke_on_key(map_unlock_codec, key_data, key=key_data, thread_id=thread_id(),
                                          reference_id=self.reference_id_generator.get_and_increment())

    def values(self, predicate=None):
        """
        Returns a list clone of the values contained in this map or values of the entries which are filtered with
        the predicate if provided.

        **Warning:
        The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and
        vice-versa.**

        :param predicate: (Predicate), predicate to filter the entries (optional).
        :return: (Sequence), a list of clone of the values contained in this map.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        if predicate:
            predicate_data = self._to_data(predicate)
            return self._encode_invoke(map_values_with_predicate_codec, predicate=predicate_data)
        else:
            return self._encode_invoke(map_values_codec)

    # internals
    def _contains_key_internal(self, key_data):
        return self._encode_invoke_on_key(map_contains_key_codec, key_data, key=key_data, thread_id=thread_id())

    def _get_internal(self, key_data):
        return self._encode_invoke_on_key(map_get_codec, key_data, key=key_data, thread_id=thread_id())

    def _get_all_internal(self, partition_to_keys, futures=None):
        if futures is None:
            futures = []
        for partition_id, key_dict in six.iteritems(partition_to_keys):
            future = self._encode_invoke_on_partition(map_get_all_codec, partition_id, keys=list(key_dict.values()))
            futures.append(future)

        def merge(f):
            return dict(itertools.chain(*f.result()))

        return combine_futures(*futures).continue_with(merge)

    def _remove_internal(self, key_data):
        return self._encode_invoke_on_key(map_remove_codec, key_data, key=key_data, thread_id=thread_id())

    def _remove_if_same_internal_(self, key_data, value_data):
        return self._encode_invoke_on_key(map_remove_if_same_codec, key_data, key=key_data, value=value_data,
                                          thread_id=thread_id())

    def _delete_internal(self, key_data):
        return self._encode_invoke_on_key(map_delete_codec, key_data, key=key_data, thread_id=thread_id())

    def _put_internal(self, key_data, value_data, ttl):
        return self._encode_invoke_on_key(map_put_codec, key_data, key=key_data, value=value_data, thread_id=thread_id(),
                                          ttl=to_millis(ttl))

    def _set_internal(self, key_data, value_data, ttl):
        return self._encode_invoke_on_key(map_set_codec, key_data, key=key_data, value=value_data, thread_id=thread_id(),
                                          ttl=to_millis(ttl))

    def _try_remove_internal(self, key_data, timeout):
        return self._encode_invoke_on_key(map_try_remove_codec, key_data, key=key_data, thread_id=thread_id(),
                                          timeout=to_millis(timeout))

    def _try_put_internal(self, key_data, value_data, timeout):
        return self._encode_invoke_on_key(map_try_put_codec, key_data, key=key_data, value=value_data,
                                          thread_id=thread_id(), timeout=to_millis(timeout))

    def _put_transient_internal(self, key_data, value_data, ttl):
        return self._encode_invoke_on_key(map_put_transient_codec, key_data, key=key_data, value=value_data,
                                          thread_id=thread_id(), ttl=to_millis(ttl))

    def _put_if_absent_internal(self, key_data, value_data, ttl):
        return self._encode_invoke_on_key(map_put_if_absent_codec, key_data, key=key_data, value=value_data,
                                          thread_id=thread_id(), ttl=to_millis(ttl))

    def _replace_if_same_internal(self, key_data, old_value_data, new_value_data):
        return self._encode_invoke_on_key(map_replace_if_same_codec, key_data, key=key_data, test_value=old_value_data,
                                          value=new_value_data, thread_id=thread_id())

    def _replace_internal(self, key_data, value_data):
        return self._encode_invoke_on_key(map_replace_codec, key_data, key=key_data, value=value_data, thread_id=thread_id())

    def _evict_internal(self, key_data):
        return self._encode_invoke_on_key(map_evict_codec, key_data, key=key_data, thread_id=thread_id())

    def _load_all_internal(self, key_data_list, replace_existing_values):
        return self._encode_invoke(map_load_given_keys_codec, keys=key_data_list, replace_existing_values=replace_existing_values)

    def _execute_on_key_internal(self, key_data, entry_processor):
        return self._encode_invoke_on_key(map_execute_on_key_codec, key_data, key=key_data,
                                          entry_processor=self._to_data(entry_processor), thread_id=thread_id())


class MapFeatNearCache(Map):
    """
    Map proxy implementation featuring Near Cache
    """
    def __init__(self, client, service_name, name):
        super(MapFeatNearCache, self).__init__(client, service_name, name)
        self._invalidation_listener_id = None
        self._near_cache = client.near_cache_manager.get_or_create_near_cache(name)
        if self._near_cache.invalidate_on_change:
            self._add_near_cache_invalidation_listener()

    def clear(self):
        self._near_cache._clear()
        return super(MapFeatNearCache, self).clear()

    def evict_all(self):
        self._near_cache.clear()
        return super(MapFeatNearCache, self).evict_all()

    def load_all(self, keys=None, replace_existing_values=True):
        if keys is None and replace_existing_values:
            self._near_cache.clear()
        return super(MapFeatNearCache, self).load_all(keys, replace_existing_values)

    def _on_destroy(self):
        self._remove_near_cache_invalidation_listener()
        self._near_cache.clear()
        super(MapFeatNearCache, self)._on_destroy()

    def _add_near_cache_invalidation_listener(self):
        try:
            request = map_add_near_cache_entry_listener_codec.encode_request(self.name, EntryEventType.invalidation,
                                                                             self._is_smart)
            self._invalidation_listener_id = self._register_listener(
                request, lambda r: map_add_near_cache_entry_listener_codec.decode_response(r)['response'],
                lambda reg_id: map_remove_entry_listener_codec.encode_request(self.name, reg_id),
                lambda m: map_add_near_cache_entry_listener_codec.handle(m, self._handle_invalidation,
                                                                         self._handle_batch_invalidation))
        except:
            self.logger.severe("-----------------\n Near Cache is not initialized!!! \n-----------------")

    def _remove_near_cache_invalidation_listener(self):
        if self._invalidation_listener_id:
            self.remove_entry_listener(self._invalidation_listener_id)

    def _handle_invalidation(self, key):
        # key is always ``Data``
        # null key means near cache has to remove all entries in it.
        # see MapAddNearCacheEntryListenerMessageTask.
        if key is None:
            self._near_cache._clear()
        else:
            self._invalidate_cache(key)

    def _handle_batch_invalidation(self, keys):
        # key_list is always list of ``Data``
        for key_data in keys:
            self._invalidate_cache(key_data)

    def _invalidate_cache(self, key_data):
        self._near_cache._invalidate(key_data)

    def _invalidate_cache_batch(self, key_data_list):
        for key_data in key_data_list:
            self._near_cache._invalidate(key_data)

    # internals
    def _contains_key_internal(self, key_data):
        try:
            return self._near_cache[key_data]
        except KeyError:
            return super(MapFeatNearCache, self)._contains_key_internal(key_data)

    def _get_internal(self, key_data):
        try:
            value = self._near_cache[key_data]
            return ImmediateFuture(value)
        except KeyError:
            future = super(MapFeatNearCache, self)._get_internal(key_data)
            return future.continue_with(self._update_cache, key_data)

    def _update_cache(self, f, key_data):
        self._near_cache.__setitem__(key_data, f.result())
        return f.result()

    def _get_all_internal(self, partition_to_keys, futures=None):
        if futures is None:
            futures = []
        for key_dic in six.itervalues(partition_to_keys):
            for key in list(key_dic.keys()):
                try:
                    key_data = key_dic[key]
                    value = self._near_cache[key_data]
                    future = ImmediateFuture((key, value))
                    futures.append(future)
                    del key_dic[key]
                except KeyError:
                    pass
        return super(MapFeatNearCache, self)._get_all_internal(partition_to_keys, futures)

    def _try_remove_internal(self, key_data, timeout):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._try_remove_internal(key_data, timeout)

    def _try_put_internal(self, key_data, value_data, timeout):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._try_put_internal(key_data, value_data, timeout)

    def _set_internal(self, key_data, value_data, ttl):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._set_internal(key_data, value_data, ttl)

    def _replace_internal(self, key_data, value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._replace_internal(key_data, value_data)

    def _replace_if_same_internal(self, key_data, old_value_data, new_value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._replace_if_same_internal(key_data, old_value_data, new_value_data)

    def _remove_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._remove_internal(key_data)

    def _remove_if_same_internal_(self, key_data, value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._remove_if_same_internal_(key_data, value_data)

    def _put_transient_internal(self, key_data, value_data, ttl):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_transient_internal(key_data, value_data, ttl)

    def _put_internal(self, key_data, value_data, ttl):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_internal(key_data, value_data, ttl)

    def _put_if_absent_internal(self, key_data, value_data, ttl):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_if_absent_internal(key_data, value_data, ttl)

    def _load_all_internal(self, key_data_list, replace_existing_values):
        self._invalidate_cache_batch(key_data_list)
        return super(MapFeatNearCache, self)._load_all_internal(key_data_list, replace_existing_values)

    def _execute_on_key_internal(self, key_data, entry_processor):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._execute_on_key_internal(key_data, entry_processor)

    def _evict_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._evict_internal(key_data)

    def _delete_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._delete_internal(key_data)


def create_map_proxy(client, service_name, name, **kwargs):
    near_cache_config = client.config.near_cache_configs.get(name, None)
    if near_cache_config is None:
        return Map(client=client, service_name=service_name, name=name)
    else:
        return MapFeatNearCache(client=client, service_name=service_name, name=name)




LOCK_ISLOCKED = 0x0701
LOCK_ISLOCKEDBYCURRENTTHREAD = 0x0702
LOCK_GETLOCKCOUNT = 0x0703
LOCK_GETREMAININGLEASETIME = 0x0704
LOCK_LOCK = 0x0705
LOCK_UNLOCK = 0x0706
LOCK_FORCEUNLOCK = 0x0707
LOCK_TRYLOCK = 0x0708




TRANSACTIONALMAP_CONTAINSKEY = 0x1001
TRANSACTIONALMAP_GET = 0x1002
TRANSACTIONALMAP_GETFORUPDATE = 0x1003
TRANSACTIONALMAP_SIZE = 0x1004
TRANSACTIONALMAP_ISEMPTY = 0x1005
TRANSACTIONALMAP_PUT = 0x1006
TRANSACTIONALMAP_SET = 0x1007
TRANSACTIONALMAP_PUTIFABSENT = 0x1008
TRANSACTIONALMAP_REPLACE = 0x1009
TRANSACTIONALMAP_REPLACEIFSAME = 0x100a
TRANSACTIONALMAP_REMOVE = 0x100b
TRANSACTIONALMAP_DELETE = 0x100c
TRANSACTIONALMAP_REMOVEIFSAME = 0x100d
TRANSACTIONALMAP_KEYSET = 0x100e
TRANSACTIONALMAP_KEYSETWITHPREDICATE = 0x100f
TRANSACTIONALMAP_VALUES = 0x1010
TRANSACTIONALMAP_VALUESWITHPREDICATE = 0x1011



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.codec.transaction_message_type import *

REQUEST_TYPE = TRANSACTION_ROLLBACK
RESPONSE_TYPE = 100
RETRYABLE = False


def calculate_size(transaction_id, thread_id):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(transaction_id)
    data_size += LONG_SIZE_IN_BYTES
    return data_size


def encode_request(transaction_id, thread_id):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(transaction_id, thread_id))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(transaction_id)
    client_message.append_long(thread_id)
    client_message.update_frame_length()
    return client_message


# Empty decode_response(client_message), this message has no parameters to decode



"""
User API for Serialization.
"""


class ObjectDataOutput(object):
    """
    ObjectDataOutput provides an interface to convert any of primitive types or arrays of them to series of bytes and
    write them on a stream.
    """
    def write_from(self, buff, offset=None, length=None):
        """
        Writes the content of the buffer to this output stream.

        :param buff: input buffer.
        :param offset: (int), offset of the buffer where copy begin (optional).
        :param length: (int), length of data to be copied from the offset into stream (optional).
        """
        raise NotImplementedError()

    def write_boolean(self, val):
        """
        Writes a bool value to this output stream. Single byte value 1 represent True, 0 represent False

        :param val: (bool), the bool to be written.
        """
        raise NotImplementedError()

    def write_byte(self, val):
        """
        Writes a byte value to this output stream.

        :param val: (byte), the byte value to be written.
        """
        raise NotImplementedError()

    def write_short(self, val):
        """
        Writes a short value to this output stream.

        :param val: (short), the short value to be written.
        """
        raise NotImplementedError()

    def write_char(self, val):
        """
        Writes a char value to this output stream.

        :param val: (char), the char value to be written.
        """
        raise NotImplementedError()

    def write_int(self, val):
        """
        Writes a int value to this output stream.

        :param val: (int), the int value to be written.
        """
        raise NotImplementedError()

    def write_long(self, val):
        """
        Writes a long value to this output stream.

        :param val: (long), the long value to be written.
        """
        raise NotImplementedError()

    def write_float(self, val):
        """
        Writes a float value to this output stream.

        :param val: (float), the float value to be written.
        """
        raise NotImplementedError()

    def write_double(self, val):
        """
        Writes a double value to this output stream.

        :param val: (double), the double value to be written.
        """
        raise NotImplementedError()

    def write_bytes(self, string):
        """
        Writes a string to this output stream.

        :param val: (str), the string to be written.
        """
        raise NotImplementedError()

    def write_chars(self, val):
        """
        Writes every character of string to this output stream.

        :param val: (str), the string to be written.
        """
        raise NotImplementedError()

    def write_utf(self, val):
        """
        Writes 2 bytes of length information and UTF-8 string to this output stream.

        :param val: (UTF-8 str), the UTF-8 string to be written.
        """
        raise NotImplementedError()

    def write_byte_array(self, val):
        """
        Writes a byte array to this output stream.

        :param val: (byte array), the byte array to be written.
        """
        raise NotImplementedError()

    def write_boolean_array(self, val):
        """
        Writes a bool array to this output stream.

        :param val: (bool array), the bool array to be written.
        """
        raise NotImplementedError()

    def write_char_array(self, val):
        """
        Writes a char array to this output stream.

        :param val: (char array), the char array to be written.
        """
        raise NotImplementedError()

    def write_int_array(self, val):
        """
        Writes a int array to this output stream.

        :param val: (int array), the int array to be written.
        """
        raise NotImplementedError()

    def write_long_array(self, val):
        """
        Writes a long array to this output stream.

        :param val: (long array), the long array to be written.
        """
        raise NotImplementedError()

    def write_double_array(self, val):
        """
        Writes a double array to this output stream.

        :param val: (double array), the double array to be written.
        """
        raise NotImplementedError()

    def write_float_array(self, val):
        """
        Writes a float array to this output stream.

        :param val: (float array), the float array to be written.
        """
        raise NotImplementedError()

    def write_short_array(self, val):
        """
        Writes a short array to this output stream.

        :param val: (short array), the short array to be written.
        """
        raise NotImplementedError()

    def write_utf_array(self, val):
        """
        Writes a UTF-8 String array to this output stream.

        :param val: (UTF-8 String array), the UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_object(self, val):
        """
        Writes an object to this output stream.

        :param val: (object), the object to be written.
        """
        raise NotImplementedError()

    def write_data(self, val):
        """
        Writes a data to this output stream.

        :param val: (Data), the data to be written.
        """
        raise NotImplementedError()

    def to_byte_array(self):
        """
        Returns a copy of internal byte array.

        :return: (byte array), the copy of internal byte array.
        """
        raise NotImplementedError()

    def get_byte_order(self):
        """
        Returns the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.

        :return: the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.
        """
        raise NotImplementedError()


class ObjectDataInput(object):
    """
    ObjectDataInput provides an interface to read bytes from a stream and reconstruct it to any of primitive types or
    arrays of them.
    """
    def read_into(self, buff, offset=None, length=None):
        """
        Reads the content of the buffer into an array of bytes.

        :param buff: input buffer.
        :param offset: (int), offset of the buffer where the read begin (optional).
        :param length: (int), length of data to be read (optional).
        :return: (byte array), the read data.
        """
        raise NotImplementedError()

    def skip_bytes(self, count):
        """
        Skips over given number of bytes from input stream.

        :param count: (long), number of bytes to be skipped.
        :return: (long), the actual number of bytes skipped.
        """
        raise NotImplementedError()

    def read_boolean(self):
        """
        Reads 1 byte from input stream and convert it to a bool value.

        :return: (bool), the bool value read.
        """
        raise NotImplementedError()

    def read_byte(self):
        """
        Reads 1 byte from input stream and returns it.

        :return: (byte), the byte value read.
        """
        raise NotImplementedError()

    def read_unsigned_byte(self):
        """
        Reads 1 byte from input stream, zero-extends it and returns.

        :return: (unsigned byte), the unsigned byte value read.
        """
        raise NotImplementedError()

    def read_short(self):
        """
        Reads 2 bytes from input stream and returns a short value.

        :return: (short), the short value read.
        """
        raise NotImplementedError()

    def read_unsigned_short(self):
        """
        Reads 2 bytes from input stream and returns an int value.

        :return: (unsigned short), the unsigned short value read.
        """
        raise NotImplementedError()

    def read_int(self):
        """
        Reads 4 bytes from input stream and returns an int value.

        :return: (int), the int value read.
        """
        raise NotImplementedError()

    def read_long(self):
        """
        Reads 8 bytes from input stream and returns a long value.

        :return: (long), the int value read.
        """
        raise NotImplementedError()

    def read_float(self):
        """
        Reads 4 bytes from input stream and returns a float value.

        :return: (float), the float value read.
        """
        raise NotImplementedError()

    def read_double(self):
        """
        Reads 8 bytes from input stream and returns a double value.

        :return: (long), the double value read.
        """
        raise NotImplementedError()

    def read_utf(self):
        """
        Reads a UTF-8 string from input stream and returns it.

        :return: (UTF-8 str), the UTF-8 string read.
        """
        raise NotImplementedError()

    def read_byte_array(self):
        """
        Reads a byte array from input stream and returns it.

        :return: (byte array), the byte array read.
        """
        raise NotImplementedError()

    def read_boolean_array(self):
        """
        Reads a bool array from input stream and returns it.

        :return: (bool array), the bool array read.
        """
        raise NotImplementedError()

    def read_char_array(self):
        """
        Reads a char array from input stream and returns it.

        :return: (char array), the char array read.
        """
        raise NotImplementedError()

    def read_int_array(self):
        """
        Reads a int array from input stream and returns it.

        :return: (int array), the int array read.
        """
        raise NotImplementedError()

    def read_long_array(self):
        """
        Reads a long array from input stream and returns it.

        :return: (long array), the long array read.
        """
        raise NotImplementedError()

    def read_double_array(self):
        """
        Reads a double array from input stream and returns it.

        :return: (double array), the double array read.
        """
        raise NotImplementedError()

    def read_float_array(self):
        """
        Reads a float array from input stream and returns it.

        :return: (float array), the float array read.
        """
        raise NotImplementedError()

    def read_short_array(self):
        """
        Reads a short array from input stream and returns it.

        :return: (short array), the short array read.
        """
        raise NotImplementedError()

    def read_utf_array(self):
        """
        Reads a UTF-8 String array from input stream and returns it.

        :return: (UTF-8 String  array), the UTF-8 String array read.
        """
        raise NotImplementedError()

    def read_object(self):
        """
        Reads a object from input stream and returns it.

        :return: (object), the object read.
        """
        raise NotImplementedError()

    def read_data(self):
        """
        Reads a data from input stream and returns it.

        :return: (Data), the data read.
        """
        raise NotImplementedError()

    def get_byte_order(self):
        """
        Returns the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.

        :return: the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.
        """
        raise NotImplementedError()

    def position(self):
        """
        Returns current position in buffer.

        :return: current position in buffer.
        """
        raise NotImplementedError()

    def size(self):
        """
        Returns size of buffer.

        :return: size of buffer.
        """
        raise NotImplementedError()


class IdentifiedDataSerializable(object):
    """
    IdentifiedDataSerializable is an alternative serialization method to Python pickle, which also avoids reflection
    during de-serialization. Each IdentifiedDataSerializable is created by a registered DataSerializableFactory.
    """
    def write_data(self, object_data_output):
        """
        Writes object fields to output stream.

        :param object_data_output: (:class:`hazelcast.serialization.api.ObjectDataOutput`), the output.
        """
        raise NotImplementedError("read_data must be implemented to serialize this IdentifiedDataSerializable")

    def read_data(self, object_data_input):
        """
        Reads fields from the input stream.

        :param object_data_input: (:class:`hazelcast.serialization.api.ObjectDataInput`), the input.
        """
        raise NotImplementedError("read_data must be implemented to deserialize this IdentifiedDataSerializable")

    def get_factory_id(self):
        """
        Returns DataSerializableFactory factory id for this class.

        :return: (int), the factory id.
        """
        raise NotImplementedError("This method must return the factory ID for this IdentifiedDataSerializable")

    def get_class_id(self):
        """
        Returns type identifier for this class. Id should be unique per DataSerializableFactory.

        :return: (int), the type id.
        """
        raise NotImplementedError("This method must return the class ID for this IdentifiedDataSerializable")


class Portable(object):
    """
    Portable provides an alternative serialization method. Instead of relying on reflection, each Portable is created by
    a registered PortableFactory. Portable serialization has the following advantages:

        * Support multiversion of the same object type.
        * Fetching individual fields without having to rely on reflection.
        * Querying and indexing support without de-serialization and/or reflection.
    """
    def write_portable(self, writer):
        """
        Serialize this portable object using given PortableWriter.

        :param writer: :class:`hazelcast.serialization.api.PortableWriter`, the PortableWriter.
        """
        raise NotImplementedError()

    def read_portable(self, reader):
        """
        Read portable fields using PortableReader.

        :param reader: :class:`hazelcast.serialization.api.PortableReader`, the PortableReader.
        """
        raise NotImplementedError()

    def get_factory_id(self):
        """
        Returns PortableFactory id for this portable class

        :return: (int), the factory id.
        """
        raise NotImplementedError()

    def get_class_id(self):
        """
        Returns class identifier for this portable class. Class id should be unique per PortableFactory.

        :return: (int), the class id.
        """
        raise NotImplementedError()


class StreamSerializer(object):
    """
    A base class for custom serialization. User can register custom serializer like following:
        >>> class CustomSerializer(StreamSerializer):
        >>>     def __init__(self):
        >>>         super(CustomSerializer, self).__init__()
        >>>
        >>>     def read(self, inp):
        >>>         pass
        >>>
        >>>     def write(self, out, obj):
        >>>         pass
        >>>
        >>>     def get_type_id(self):
        >>>         pass
        >>>
        >>>     def destroy(self):
        >>>         pass

    """
    def write(self, out, obj):
        """
        Writes object to ObjectDataOutput

        :param out: (:class:`hazelcast.serialization.api.ObjectDataOutput`), stream that object will be written to.
        :param obj: (object), the object to be written.
        """
        raise NotImplementedError("write method must be implemented")

    def read(self, inp):
        """
        Reads object from objectDataInputStream

        :param inp: (:class:`hazelcast.serialization.api.ObjectDataInput`) stream that object will read from.
        :return: (object), the read object.
        """
        raise NotImplementedError("write method must be implemented")

    def get_type_id(self):
        """
        Returns typeId of serializer.
        :return: (int), typeId of serializer.
        """
        raise NotImplementedError("get_type_id must be implemented")

    def destroy(self):
        """
        Called when instance is shutting down. It can be used to clear used resources.
        """
        raise NotImplementedError()


class PortableReader(object):
    """
    Provides a mean of reading portable fields from binary in form of Python primitives and arrays of these primitives,
    nested portable fields and array of portable fields.
    """
    def get_version(self):
        """
        Returns the global version of portable classes.

        :return: (int), global version of portable classes.
        """
        raise NotImplementedError()

    def has_field(self, field_name):
        """
        Determine whether the field name exists in this portable class or not.

        :param field_name: (str), name of the field (does not support nested paths).
        :return: (bool), ``true`` if the field name exists in class, ``false`` otherwise.
        """
        raise NotImplementedError()

    def get_field_names(self):
        """
        Returns the set of field names on this portable class.

        :return: (Set), set of field names on this portable class.
        """
        raise NotImplementedError()

    def get_field_type(self, field_name):
        """
        Returns the field type of given field name.

        :param field_name: (str), name of the field.
        :return: the field type.
        """
        raise NotImplementedError()

    def get_field_class_id(self, field_name):
        """
        Returns the class id of given field.

        :param field_name: (str), name of the field.
        :return: (int), class id of given field.
        """
        raise NotImplementedError()

    def read_int(self, field_name):
        """
        Reads a primitive int.

        :param field_name: (str), name of the field.
        :return: (int), the int value read.
        """
        raise NotImplementedError()

    def read_long(self, field_name):
        """
        Reads a primitive long.

        :param field_name: (str), name of the field.
        :return: (long), the long value read.
        """
        raise NotImplementedError()

    def read_utf(self, field_name):
        """
        Reads a UTF-8 String.

        :param field_name: (str), name of the field.
        :return: (UTF-8 str), the UTF-8 String read.
        """
        raise NotImplementedError()

    def read_boolean(self, field_name):
        """
        Reads a primitive bool.

        :param field_name: (str), name of the field.
        :return: (bool), the bool value read.
        """
        raise NotImplementedError()

    def read_byte(self, field_name):
        """
        Reads a primitive byte.

        :param field_name: (str), name of the field.
        :return: (byte), the byte value read.
        """
        raise NotImplementedError()

    def read_char(self, field_name):
        """
        Reads a primitive char.

        :param field_name: (str), name of the field.
        :return: (char), the char value read.
        """
        raise NotImplementedError()

    def read_double(self, field_name):
        """
        Reads a primitive double.

        :param field_name: (str), name of the field.
        :return: (double), the double value read.
        """
        raise NotImplementedError()

    def read_float(self, field_name):
        """
        Reads a primitive float.

        :param field_name: (str), name of the field.
        :return: (float), the float value read.
        """
        raise NotImplementedError()

    def read_short(self, field_name):
        """
        Reads a primitive short.

        :param field_name: (str), name of the field.
        :return: (short), the short value read.
        """
        raise NotImplementedError()

    def read_portable(self, field_name):
        """
        Reads a portable.

        :param field_name: (str), name of the field.
        :return: (:class:`hazelcast.serialization.api.Portable`, the portable read.
        """
        raise NotImplementedError()

    def read_byte_array(self, field_name):
        """
        Reads a primitive byte array.

        :param field_name: (str), name of the field.
        :return: (byte array), the byte array read.
        """
        raise NotImplementedError()

    def read_boolean_array(self, field_name):
        """
        Reads a primitive bool array.

        :param field_name: (str), name of the field.
        :return: (bool array), the bool array read.
        """
        raise NotImplementedError()

    def read_char_array(self, field_name):
        """
        Reads a primitive char array.

        :param field_name: (str), name of the field.
        :return: (char array), the char array read.
        """
        raise NotImplementedError()

    def read_int_array(self, field_name):
        """
        Reads a primitive int array.

        :param field_name: (str), name of the field.
        :return: (int array), the int array read.
        """
        raise NotImplementedError()

    def read_long_array(self, field_name):
        """
        Reads a primitive long array.

        :param field_name: (str), name of the field.
        :return: (long array), the long array read.
        """
        raise NotImplementedError()

    def read_double_array(self, field_name):
        """
        Reads a primitive double array.

        :param field_name: (str), name of the field.
        :return: (double array), the double array read.
        """
        raise NotImplementedError()

    def read_float_array(self, field_name):
        """
        Reads a primitive float array.

        :param field_name: (str), name of the field.
        :return: (float array), the float array read.
        """
        raise NotImplementedError()

    def read_short_array(self, field_name):
        """
        Reads a primitive short array.

        :param field_name: (str), name of the field.
        :return: (short array), the short array read.
        """
        raise NotImplementedError()

    def read_utf_array(self, field_name):
        """
        Reads a UTF-8 String array.

        :param field_name: (str), name of the field.
        :return: (UTF-8 String array), the UTF-8 String array read.
        """
        raise NotImplementedError()

    def read_portable_array(self, field_name):
        """
        Reads a portable array.

        :param field_name: (str), name of the field.
        :return: (Portable array), the portable array read.
        """
        raise NotImplementedError()

    def get_raw_data_input(self):
        """
        After reading portable fields, one can read remaining fields in old fashioned way consecutively from the end of
        stream. After get_raw_data_input() called, no data can be read.

        :return: (:class:`hazelcast.serialization.api.ObjectDataInput`), the input.
        """
        raise NotImplementedError()


class PortableWriter(object):
    """
    Provides a mean of writing portable fields to a binary in form of Python primitives and arrays of these primitives,
    nested portable fields and array of portable fields.
    """
    def write_int(self, field_name, value):
        """
        Writes a primitive int.

        :param field_name: (str), name of the field.
        :param value: (int), int value to be written.
        """
        raise NotImplementedError()

    def write_long(self, field_name, value):
        """
        Writes a primitive long.

        :param field_name: (str), name of the field.
        :param value: (long), long value to be written.
        """
        raise NotImplementedError()

    def write_utf(self, field_name, value):
        """
        Writes an UTF string.

        :param field_name: (str), name of the field.
        :param value: (UTF str), UTF string value to be written.
        """
        raise NotImplementedError()

    def write_boolean(self, field_name, value):
        """
        Writes a primitive bool.

        :param field_name: (str), name of the field.
        :param value: (bool), bool value to be written.
        """
        raise NotImplementedError()

    def write_byte(self, field_name, value):
        """
        Writes a primitive byte.

        :param field_name: (str), name of the field.
        :param value: (byte), byte value to be written.
        """
        raise NotImplementedError()

    def write_char(self, field_name, value):
        """
        Writes a primitive char.

        :param field_name: (str), name of the field.
        :param value: (char), char value to be written.
        """
        raise NotImplementedError()

    def write_double(self, field_name, value):
        """
        Writes a primitive double.

        :param field_name: (str), name of the field.
        :param value: (Dobule), double value to be written.
        """
        raise NotImplementedError()

    def write_float(self, field_name, value):
        """
        Writes a primitive float.

        :param field_name: (str), name of the field.
        :param value: (float), float value to be written.
        """
        raise NotImplementedError()

    def write_short(self, field_name, value):
        """
        Writes a primitive short.

        :param field_name: (str), name of the field.
        :param value: (short), short value to be written.
        """
        raise NotImplementedError()

    def write_portable(self, field_name, portable):
        """
        Writes a Portabl

        :param field_name: (str), name of the field.
        :param portable: (:class:`hazelcast.serialization.api.Portable`, portable to be written.
        """
        raise NotImplementedError()

    def write_null_portable(self, field_name, factory_id, class_id):
        """
        To write a null portable value, user needs to provide class and factoryIds of related class.

        :param field_name: (str), name of the field.
        :param factory_id: (int), factory id of related portable class.
        :param class_id: (int), class id of related portable class.
        """
        raise NotImplementedError()

    def write_byte_array(self, field_name, values):
        """
        Writes a primitive byte array.

        :param field_name: (str), name of the field.
        :param values: (byte array), byte array to be written.
        """
        raise NotImplementedError()

    def write_boolean_array(self, field_name, values):
        """
        Writes a primitive bool array.

        :param field_name: (str), name of the field.
        :param values: (bool array), bool array to be written.
        """
        raise NotImplementedError()

    def write_char_array(self, field_name, values):
        """
        Writes a primitive char array.

        :param field_name: (str), name of the field.
        :param values: (char array), char array to be written.
        """
        raise NotImplementedError()

    def write_int_array(self, field_name, values):
        """
        Writes a primitive int array.

        :param field_name: (str), name of the field.
        :param values: (int array), int array to be written.
        """
        raise NotImplementedError()

    def write_long_array(self, field_name, values):
        """
        Writes a primitive long array.

        :param field_name: (str), name of the field.
        :param values: (long array), long array to be written.
        """
        raise NotImplementedError()

    def write_double_array(self, field_name, values):
        """
        Writes a primitive double array.

        :param field_name: (str), name of the field.
        :param values: (double array), double array to be written.
        """
        raise NotImplementedError()

    def write_float_array(self, field_name, values):
        """
        Writes a primitive float array.

        :param field_name: (str), name of the field.
        :param values: (float array), float array to be written.
        """
        raise NotImplementedError()

    def write_short_array(self, field_name, values):
        """
        Writes a primitive short array.

        :param field_name: (str), name of the field.
        :param values: (short array), short array to be written.
        """
        raise NotImplementedError()

    def write_utf_array(self, field_name, values):
        """
        Writes a UTF-8 String array.

        :param field_name: (str), name of the field.
        :param values: (UTF-8 String array), UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_portable_array(self, field_name, values):
        """
        Writes a portable array.

        :param field_name: (str), name of the field.
        :param values: (Portable array), portable array to be written.
        """
        raise NotImplementedError()

    def get_raw_data_output(self):
        """
        After writing portable fields, one can write remaining fields in old fashioned way consecutively at the end of
        stream. After get_raw_data_output() called, no data can be written.

        :return: (:class:`hazelcast.serialization.api.ObjectDataOutput`), the output.
        """
        raise NotImplementedError()

